import logging
import os
import sys
from logging.handlers import TimedRotatingFileHandler

__all__ = ['set_standard_logging_config', 'init_root_logger']

def suppress_common_packages():
    """Suppress INFO logs from common packages to reduce clutter"""
    # setting warning level to various packages to avoid additional logs
    packages = ('botocore', 'boto3', 'requests', 'urllib3', 'rasterio', 'shapely', 'paramiko')
    for p in packages:
        p_logger = logging.getLogger(p)
        p_logger.setLevel(logging.WARNING)


def get_fmt_string(code: str = None, env: str = None) -> str:
    """Creates common format string for logging

    Args:
        code: an optional flight code to include
        env: an optional environment to include

    Returns:
        a format string
    """
    format_str = "[{asctime}][{levelname}][{name}]: {message}"

    if code:
        format_str = "[{}]{}".format(code, format_str)

    if env:
        format_str = "[{}]{}".format(env, format_str)

    return format_str


def set_standard_logging_config(*, level=logging.INFO, stdout: bool = True, stderr: bool = False, logfile: str = None,
                                code: str = None, env: str = None, suppress_packages: bool = True):
    """Sets a standard logging context

    Args:
        level: the logging level of the root logger
        stdout: if true will print logs to stdout
        stderr: if true will print logs to stderr
        logfile: an optional path to write logs to
        code: an optional code to include in the logging output
        env: an optional environment to include in the logging output
        suppress_packages: if true will suppress info messages from common packages
    """

    if env is None and 'IA_ENV' in os.environ:
        env = os.environ['IA_ENV']

    if suppress_packages:
        suppress_common_packages()

    if 'IA_LOGGING_FMT' in os.environ and os.environ['IA_LOGGING_FMT'] != '':
        fmt_string = os.environ['IA_LOGGING_FMT']
    else:
        fmt_string = get_fmt_string(code=code, env=env)

    handlers = list()

    if stdout:
        handlers.append(logging.StreamHandler(sys.stdout))

    if stderr:
        handlers.append(logging.StreamHandler(sys.stderr))

    if logfile:
        handlers.append(logging.FileHandler(logfile))

    if 'IA_LOGFILE' in os.environ and os.environ['IA_LOGFILE'] != '':
        handlers.append(logging.FileHandler(os.environ['IA_LOGFILE']))

    logging.basicConfig(level=level, handlers=handlers, format=fmt_string, style='{', force=True)


def init_root_logger(
    logfile=None, code=None, env=None, logging_level=logging.INFO, log_rotate=False, daily_log_file=None
):
    packages = (
        "botocore",
        "boto3",
        "requests",
        "urllib3",
        "rasterio",
        "shapely",
        "paramiko",
        "fiona",
        "filelock",
        "s3transfer",
    )
    for p in packages:
        logger = logging.getLogger(p)
        logger.setLevel(logging.WARNING)

    # resets all loggers handlers to escape from duplicate handlers
    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]

    for l in loggers:
        if l.handlers:
            l.handlers.clear()

    format_str = get_fmt_string(code=code, env=env)

    handlers = []
    handlers.append(logging.StreamHandler(sys.stdout))

    if logfile:
        handlers.append(logging.FileHandler(logfile))

    if log_rotate:
        # This piece of code will create a log file but the log will be moved to a new log file
        # named log-daily-{machine_name}.txt.{previous_date}
        # when the current day ends at midnight.

        handler = TimedRotatingFileHandler(daily_log_file, when="midnight", backupCount=1)
        handlers.append(handler)

    root_logger = logging.getLogger()

    # imitating `force=True` flag in `logging.basicConfig`
    root_logger.handlers.clear()

    for handler in handlers:
        handler.setFormatter(logging.Formatter(format_str, style="{"))
        handler.setLevel(logging_level)
        root_logger.addHandler(handler)



"""Processing Zone service API wrapper class."""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Union, get_type_hints, Optional
from urllib.parse import urlencode
from uuid import UUID

from intelinair_utils.api_wrapper import ApiWrapper


@dataclass
class ProcessingZone:  # pylint: disable=too-many-instance-attributes
    """Processing Zone dataclass.
    
    Used for type hinting and data validation.
    """
    field_token: UUID
    season_id: int
    id: int = None  # pylint: disable=invalid-name
    token: UUID = None
    version: int = None
    date_created: datetime = None
    last_updated: datetime = None
    dates_active: Dict[str, Any] = None
    is_active: bool = None
    zone_type: str = None
    operation_date: datetime = None
    operation_type: str = None
    geometry: Dict[str, Any] = None
    parent: int = None
    extras: Dict[Any, Any] = None

    @classmethod
    def get_instance(cls, data: Dict[str, Any]) -> 'ProcessingZone':
        """Get ProcessingZone instance from dictionary.

        Args:
            data: Dictionary containing data.

        Returns:
            ProcessingZone instance.
        """
        class_properties = get_type_hints(cls)
        deserialized_properties = {}
        for attr, type_ in class_properties.items():
            try:
                if type_ == datetime:
                    if isinstance(data[attr], str):
                        value = datetime.fromisoformat(data[attr])
                    elif isinstance(data[attr], (datetime, type(None))):
                        value = data[attr]
                    else:
                        raise TypeError(f'{attr} must be a datetime or string')
                elif type_ == UUID:
                    if isinstance(data[attr], str):
                        value = UUID(data[attr])
                    elif isinstance(data[attr], (UUID, type(None))):
                        value = data[attr]
                    else:
                        raise TypeError(f'{attr} must be a UUID or string')
                else:
                    value = data[attr]
            except KeyError:
                continue
            deserialized_properties[attr] = value
        return cls(**deserialized_properties)

    def _serialize(self, value: Any) -> Any:
        """Serialize value to JSON compatible type.

        Args:
            value (Any): Value to serialize.

        Returns:
            Any: Serialized value.
        """
        if isinstance(value, datetime):
            return value.isoformat()
        if isinstance(value, UUID):
            return str(value)
        return value

    def to_dict(self) -> Dict[str, Any]:
        """Convert dataclass to dictionary.
        """
        return {
            key: self._serialize(value)
            for key, value in self.__dict__.items()
            if value is not None
        }

    def patched_dict(self, **kwargs) -> Dict[str, Any]:
        """Convert dataclass to dictionary and patch with kwargs.
        """
        data = self.to_dict()
        for key, value in kwargs.items():
            data[key] = self._serialize(value)
        return data


class ProcessingZoneServiceApiError(Exception):
    """Processing Zone service API error class."""
    error_code: str = None
    status_code: int = None

    def __init__(self, message: str, error_code: str = None, status_code: int = None):
        super().__init__(message)
        self.message = message
        self.error_code = error_code
        self.status_code = status_code

    def __str__(self) -> str:
        return f'[HTTP {self.status_code}] {self.error_code} ({self.message})'

    def __repr__(self) -> str:
        return f'ProcessingZoneServiceApiError({self.status_code}, {self.error_code}, {self.message})'


class ProcessingZoneServiceApi(ApiWrapper):
    """Processing Zone service API class."""

    PROCESSING_ZONE_SERVICE_URLS = {
        "bellflower": "https://processing-zone-service.bellflower.int.intelinair.dev/api",
        "prod": "https://processing-zone-service.prod.int.intelinair.dev/api",
        "platform": "https://processing-zone-service.platform.int.intelinair.dev/api",
        "release": "https://processing-zone-service.release.int.intelinair.dev/api"
    }

    def __init__(self, environment: str, config_path: Optional[str] = None):
        super().__init__(environment, self._get_api_urls(), config_path=config_path)

    def _get_api_urls(self):
        """Overridden method for getting planting service API gateway URLs."""
        return self.PROCESSING_ZONE_SERVICE_URLS

    def _make_request(self, http_method, *args, **kwargs):
        response = super()._make_request(http_method, *args, **kwargs)
        if response.status_code < 200 or response.status_code >= 300:
            error_response = response.json()
            raise ProcessingZoneServiceApiError(error_response.get('Message'),
                                                error_response.get('Code'),
                                                response.status_code)
        return response

    def get_zones_for_date(
        self,
        field_token: str,
        season_id: Union[int, str],
        flight_date: datetime
    ) -> List[ProcessingZone]:
        """Get processing zones for a given date.

        Retrieves the active processing zones for a given date. In scenarios
        where there are replant zones, the replant zones will be returned
        only when the flight date is after the replant date.

        Args:
            field_token: the token of the field
            season_id: Season ID.
            flight_date: Flight date.
        
        Returns:
            List[ProcessingZone]: List of processing zones.
        """
        query_params = urlencode({
            'flight_date': flight_date.isoformat(),
            'season_id': season_id,
            'field_token': field_token
        })
        url = self.url_template.format(f"zones-for-date?{query_params}")
        response = self._make_request(self.session.get, url)
        zones = [ProcessingZone.get_instance(zone) for zone in response.json().get('zones', [])]
        return zones

    def get_latest_zones(
        self,
        field_token: str,
        season_id: Union[int, str]
    ) -> List[ProcessingZone]:
        """Get latest processing zones for a given field and season.

        Args:
            field_token (str): Field token.
            season_id (Union[int, str]): Season ID.

        Returns:
            List[ProcessingZone]: List of processing zones.
        """
        query_params = urlencode({
            'season_id': season_id,
            'field_token': field_token
        })
        url = self.url_template.format(f"zones/latest?{query_params}")
        response = self._make_request(self.session.get, url)
        zones = [ProcessingZone.get_instance(zone) for zone in response.json().get('zones', [])]
        return zones

    def get_latest_zones_by_type(
        self,
        field_token: str,
        season_id: Union[int, str],
        zone_type: str
    ) -> List[ProcessingZone]:
        """Get latest processing zones for a given field and season of a given type.

        Args:
            field_token: Field token.
            season_id: Season ID.
            zone_type: the type of the zone

        Returns:
            List[ProcessingZone]: List of processing zones.
        """
        query_params = urlencode({
            'season_id': season_id,
            'field_token': field_token,
            'zone_type': zone_type
        })
        url = self.url_template.format(f"zones/latest?{query_params}")
        response = self._make_request(self.session.get, url)
        zones = [ProcessingZone.get_instance(zone) for zone in response.json().get('zones', [])]
        return zones

    def get_zone_by_token(self, zone_token: str) -> Optional[ProcessingZone]:
        """Get processing zone for given zone token.
        Args:
            zone_token: Processing Zne token.

        Returns:
            List of processing zones.
        """
        query_params = urlencode({
            'token': zone_token
        })
        url = self.url_template.format(f"zones?{query_params}")
        response = self._make_request(self.session.get, url)
        zones = response.json().get('zones', [])
        if len(zones) == 0:
            return None
        # take last zone in case there are multiple zones with same token (revision)
        zones.sort(key=lambda x: x['version'], reverse=False)
        zone = ProcessingZone.get_instance(zones.pop())
        return zone

    def revise(
        self,
        field_token: UUID,
        season_id: int,
        zones: List[ProcessingZone]
    ) -> List[ProcessingZone]:
        """Revise processing zones for the given field.
        
        Args:
            field_token: Field token.
            season_id: Season ID.
            zones: List of processing zones.

        Returns:
            List of revised processing zones.
        """
        field_season = {
            'season_id': season_id,
            'field_token': field_token
        }
        url = self.url_template.format(f"zones")
        message_body = {
            'field_token': str(field_token),
            'season_id': season_id,
            'zones': [zone.patched_dict(**field_season) for zone in zones],
        }
        response = self._make_request(self.session.put, url, json=message_body)
        zones = [ProcessingZone.get_instance(zone) for zone in response.json().get('zones', [])]
        return zones

    def create(
        self,
        field_token: UUID,
        season_id: int,
        zones: List[ProcessingZone]
    ) -> List[ProcessingZone]:
        """Create processing zones for a given date.

        Args:
            field_token: Field token.
            season_id: Season ID.
            zones: List of processing zones.

        Returns:
            List of created processing zones.
        """
        field_season = {
            'season_id': season_id,
            'field_token': field_token
        }
        url = self.url_template.format("zones")
        data = [zone.patched_dict(**field_season) for zone in zones]
        response = self._make_request(self.session.post, url, json=data)
        return [ProcessingZone.get_instance(zone) for zone in response.json().get('zones', [])]



"""
Provides a common API for working with remote storage services.

Currently this abstracts out both uploading and downloading of
files & folders from both SFTP and S3. Permissions for SFTP
access are stored within the class itself with for S3 the
credentials need to be exposed externally. See the `Boto3 docs`_
for how to configure this appropriately.

.. _`Boto3 docs`: http://boto3.readthedocs.io/en/latest/guide/configuration.html#configuring-credentials

Example code::

   >>> Storage().download('sftp://Visual Data/', '.')
   >>> Storage().download('s3://flight-ops/VisualData/', '.')
"""

import ftplib
import logging
import os
import subprocess
import time
from abc import ABCMeta, abstractmethod
from typing import *
from urllib.parse import urlparse

import boto3

from intelinair_utils.storage_utils import get_files, path_split_all

logger = logging.getLogger("storage")


def get_backends() -> List[str]:
    """List available storage backends"""
    return list(storage_backends)


class Storage:
    """
    Primary class for working with remote files.

    The protocol backend is parsed out of the URL string and as such this class
    hides the protocol-specific implementations from the user.
    """

    def __init__(self) -> None:
        self.backends = {}  # type: Dict[str, StorageBackend]
        for n, backend in storage_backends.items():
            # Try to add all possible storage backends ignoring ones that
            # error.
            try:
                self.backends[n] = backend()
            except Exception as e:
                logger.warning("Failed to initialize storage backend '{}'" "".format(n))

    def _get_backend(self, url: str) -> "StorageBackend":
        """Returns if a URL is valid.

        Relies on a Backend for a given scheme to exist.
        """
        parsed_url = urlparse(url)
        if parsed_url.scheme not in storage_backends:
            raise ValueError("Scheme must be one of {}".format(list(storage_backends.keys())))
        else:
            return self.backends[parsed_url.scheme]

    def download(self, remote_url: str, local_path: Optional[str] = None, recursive: bool = False):
        """Download a file or the contents of a directory.

        If the remote_url is a directory, the contents of the directory will
        be downloaded into the specified local directory. If the remote_url is
        a file, it will be downloaded with the same filename into the
        specified local directory.
        """
        self._get_backend(remote_url).download(remote_url, local_path, recursive=recursive)

    def upload(self, local_path: str, remote_url: str):
        """Upload a file or the contents of a directory.

        If the local_path is a directory, the contents of the directory will
        be uploaded into the specified remote directory. If the local_path is
        a file, it will be uploaded with the same filename into the
        specified remote directory.
        """
        self._get_backend(remote_url).upload(local_path, remote_url)

    def remove(self, remote_url: str, recursive: bool = False):
        """Remove a file or the contents of a directory.

        If the remote_url is a directory, the contents of the directory will
        be removed recursively. If the remote_url is
        a file it will be removed from s3.
        """
        self._get_backend(remote_url).remove(remote_url, recursive=recursive)

    def exists(self, remote_url: str):
        """Checks if a remote url exists in S3."""
        return StorageBackendS3Cli().exists(remote_url)

    @staticmethod
    def exponential_backoff(function: Callable, *args):
        """
        Run a function using an exponential backoff during failure.

        Failure is defined here as the function raising any kind of exception.
        """
        tries = 3  # Number of times to attempt function()
        delay = 1  # Base amount of delay (seconds)
        backoff = 2  # Multiplier for determining delay after each failure
        for i in range(tries):
            try:
                return function(*args)
            except:
                logger.exception("Failed running function '{}', trying again after {}s.".format(function, delay))
                time.sleep(delay)
                delay *= backoff
        else:
            raise Exception("Failed to run '{}' {} times, aborting.".format(function, tries))


class StorageBackend(metaclass=ABCMeta):
    """Abstract class defining the interface for storage backends."""

    def __init__(self):
        pass

    @abstractmethod
    def download(self, remote_url: str, local_path: str, recursive: bool = False) -> None:
        pass

    @abstractmethod
    def upload(self, local_path: str, remote_url: str) -> None:
        pass

    @abstractmethod
    def remove(self, remote_url: str, recursive: bool = False) -> None:
        pass


class StorageBackendFtp(StorageBackend):
    """
    Implementation of an FTP backend for the Storage API.

    Assumes that explicit TLS is supported by the server (also known as FTPS).

    All login credentials are embedded in this library and no external
    configuration is necessary. Additionally this also means that only files on
    the IntelinAir FTP server can be accessed.
    """

    def __init__(self) -> None:
        self.ftp_uri = "ftp.intelinair.com"
        self.ftp_login = "bizdev@intelinair.com"
        self.ftp_pass = "FTP1simple"

    def _connect(self) -> None:
        # Log in to the FTP server and switch to a secure connection
        self.ftps = ftplib.FTP_TLS(self.ftp_uri)
        self.ftps.auth()
        self.ftps.login(self.ftp_login, self.ftp_pass)
        self.ftps.prot_p()

    def _disconnect(self) -> None:
        try:
            self.ftps.quit()
        except ftplib.all_errors as e:
            logger.exception("Exception during disconnect")

    def download(self, remote_uri: str, local_path: str, recursive: bool = False) -> None:
        """
        Fetches all files within the specified folder (relative to the root
        directory) from the IntelinAir FTP server and download to the specified
        directory. This replaces any content that was in that directory
        already.
        """
        assert recursive is False, "recursive downloading has not been tested for FTP"

        self._connect()

        # Find all the files in the desired directory
        self.ftps.cwd(remote_uri)
        files = self.ftps.nlst()
        logger.info('Found {} files in "{}"'.format(len(files), remote_uri))

        # Make sure the output directory exists
        try:
            os.makedirs(local_path)
        except FileExistsError:
            pass

        # And is empty
        # TODO

        # Copy all files to our local directory
        i = 1
        for filename in files:
            local_filename = os.path.join(local_path, filename)
            while True:
                try:
                    logger.info("{}/{} {} <- {}".format(i, len(files), local_filename, filename))

                    # Try to download the file
                    f = open(local_filename, "wb")
                    self.ftps.retrbinary("RETR {}".format(filename), f.write)
                    f.close()

                    # Verify file size of the downloaded file
                    filesize = int(self.ftps.sendcmd("SIZE {}".format(filename)).split(" ")[1])
                    stat = os.stat(local_filename)
                    if filesize != stat.st_size:
                        logger.error(
                            "Downloaded filesize ({}) doesn't match remote size ({}), retrying download.".format(
                                filesize, stat.st_size
                            )
                        )
                        os.remove(local_filename)
                        continue

                    # Stop retrying
                    break
                except ftplib.all_errors:
                    logger.exception("Failed to download '{}', trying again.".format(filename))

                    # Reconnect and go back to our data directory
                    self._disconnect()
                    self._connect()
                    self.ftps.cwd(remote_uri)
            i += 1

        self._disconnect()

    def upload(self, local_path: str, remote_uri: str) -> None:
        """
        Uploads all contents of the specified folder to the IntelinAir FTP
        server to the specified directory (remote_uri is assumed to be an
        absolute path even if the initial '/' is missing). This replaces any
        content that was in that directory already. Als all subfolder structure
        is preserved during uploading.
        """

        # Find all input files
        if os.path.isdir(local_path):
            files = get_files(local_path)
            logger.info("Found {} files in '{}'".format(len(files), local_path))
        elif os.path.isfile(local_path):
            logger.info("Found 1 file '{}'".format(local_path))
            files = [local_path]
            local_path = os.path.split(local_path)[0]
        else:
            raise Exception("ERROR: Non-file/non-directory path specified")

        # Make sure the remote_uri starts with a forward-slash
        if not remote_uri.startswith("/"):
            remote_uri = "/" + remote_uri

        # Make sure we're connected to the FTP server
        self._connect()

        # Copy all files to the remote directory
        i = 1
        for full_filename in files:
            filename = os.path.split(full_filename)[1]
            local_relpath = os.path.relpath(full_filename, local_path)
            remote_relpath = "/".join(path_split_all(local_relpath))

            dest_name = "/".join(
                [remote_uri, remote_relpath]
            )  # Don't use os.path because FTP server always uses '/' as directory separators
            logger.info("{}/{} {} -> {}".format(i, len(files), full_filename, dest_name))

            remote_folder = "/".join(dest_name.split("/")[:-1])
            self._mkdir(remote_folder)
            f = open(full_filename, "rb")
            self.ftps.storbinary("STOR {}".format(filename), f)
            f.close()
            i += 1

        self._disconnect()

    def _mkdir(self, path: str) -> None:
        """Create the given path on the FTP server"""
        dest_path = path.split("/")
        for i in range(0, len(dest_path)):
            curr_path = "/" + "/".join(dest_path[0: i + 1])
            try:
                self.ftps.cwd(curr_path)
            except ftplib.all_errors as e:
                response_code = str(e).split(" ")[0]
                if response_code == "550":
                    self.ftps.mkd(curr_path)
                    self.ftps.cwd(curr_path)
                    logger.debug("Making directory '{}'".format("/".join(dest_path[:i])))
                else:
                    logger.exception("Failed to set current working directory")

    def remove(self, remote_url: str, recursive: bool = False) -> None:
        pass


class StorageBackendS3Cli(StorageBackend):
    """
    Implementation of an Amazon S3 backend using aws cli for performance.

    This assumes that aws cli is installed and configured with the correct
    keys and default region.

    apt-get awscli

    aws configure
    key:
    code:
    default-region:us-east-1

    """

    def download(self, remote_url: str, local_path: str, recursive: bool = False):
        if local_path is None:
            local_path = "."

        # Split the path into the bucket name and the path within the bucket
        url = urlparse(remote_url)
        bucket_name = url.netloc
        if not bucket_name:
            raise ValueError("Missing bucket name in S3 URL")

        aws_cmd_string = 'aws s3 cp "' + remote_url + '" "' + local_path + '"'
        if recursive:
            aws_cmd_string = aws_cmd_string + " --recursive"

        retries_left = 5
        for _ in range(retries_left):
            logger.info("Running aws commandline like: {}".format(aws_cmd_string))
            # run aws command
            p = subprocess.Popen(aws_cmd_string, shell=True)
            if not p.wait():  # completed successfully
                break
            else:
                logger.warning("Download failed. Retries left : {}".format(retries_left))
                retries_left -= 1
        else:
            raise Exception(f"Failed to perform download operation")

    def upload(self, local_path: str, remote_url: str):
        if local_path is None:
            local_path = "."

        if not remote_url.endswith("/"):
            remote_url = remote_url + "/"

        # Split the path into the bucket name and the path within the bucket
        url = urlparse(remote_url)
        bucket_name = url.netloc
        if not bucket_name:
            raise ValueError("Missing bucket name in S3 URL")

        aws_cmd_string = 'aws s3 cp "' + local_path + '" "' + remote_url + '"'
        if os.path.isdir(local_path):
            aws_cmd_string = aws_cmd_string + " --recursive"

        retries_left = 5
        for _ in range(retries_left):
            logger.info("Running aws commandline like: {}".format(aws_cmd_string))

            # run aws command
            p = subprocess.Popen(aws_cmd_string, shell=True)
            if not p.wait():  # completed successfully
                break
            else:
                logger.warning("Upload failed. Retries left : {}".format(retries_left))
                retries_left -= 1
        else:
            raise Exception(f"Failed to perform upload operation")

    def remove(self, remote_url: str, recursive: bool = False):
        # Split the path into the bucket name and the path within the bucket
        url = urlparse(remote_url)
        bucket_name = url.netloc
        if not bucket_name:
            raise ValueError("Missing bucket name in S3 URL")

        if recursive and not remote_url.endswith("/"):
            # add "/" in the end
            remote_url = remote_url + "/"

        aws_cmd_string = 'aws s3 rm "' + remote_url + '"'
        if recursive:
            aws_cmd_string = aws_cmd_string + " --recursive"
        logger.info("Running aws commandline like: {}".format(aws_cmd_string))

        # run aws command
        p = subprocess.Popen(aws_cmd_string, shell=True)
        p.wait()
        return

    def exists(self, remote_url: str):
        self._connect()
        url = urlparse(remote_url)
        bucket_name, prefix = url.netloc, url.path.lstrip("/")
        if not bucket_name:
            raise ValueError("Missing bucket name in S3 URL")

        result = self.s3.list_objects_v2(Bucket=bucket_name, Prefix=prefix)
        self._disconnect()
        return "Contents" in result

    def _connect(self) -> None:
        self.s3 = boto3.client("s3")

    def _disconnect(self) -> None:
        self.s3 = None


storage_backends = {"ftp": StorageBackendFtp, "s3": StorageBackendS3Cli}



from .__version__ import __version__
from .agmri_api import AgmriApi
from .alert_service_api import AlertServiceApi
from .ec2_utils import *
from .efs_utils import *
from .error_logger import ErrorLogger
from .flight_service_api import FlightServiceApi, AERIAL, EQUIPMENT, IMAGERY
from .integration_service_api import IntegrationServiceApi
from .logging_utils import *
from .ml_api import MLApi
from .os_utils import *
from .pipeline_stats_api import PipelineStatsApi
from .processing_zone_service_api import ProcessingZoneServiceApi, ProcessingZone
from .s3_utils import *
from .ssm_utils import *
from .vector_tile_service_api import VectorTileServiceApi
from .weather_api import WeatherApi
from .efs_utils import get_available_efses, get_available_efses_from_s3
try:
    from .annotations_api import AnnotationsApi
    from .planting_service_api import PlantingServiceApi, PlantingServiceOperation
    from .elevation_metadata_service_api import ElevationMetadataServiceApi
except ImportError:
    import logging
    logger = logging.getLogger('intelinair_utils')
    logger.warning("The optional dependency 'shapely' is missing, which means certain functionalities "
                   "(such as AnnotationsApi) will be unavailable. To enable these features, install 'shapely' "
                   "by running: pip install intelinair_utils[shapely].")



import datetime
from typing import Dict, List, Optional, Tuple

import pytz
import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from intelinair_utils.agmri_api import AgmriApi


class WeatherApi:
    """Special class for working getting weather data."""

    DAILY_REQUEST_KEY = 'historicalDaily'
    HOURLY_REQUEST_KEY = 'historicalHourly'
    DAY_KEY = 'day'
    HOUR_KEY = 'hour'

    def __init__(self, api: AgmriApi, number_of_retries: int = 3) -> None:
        """Class responsible for communication with weather service."""
        self._api = api
        self._weather_url_fmt = self._get_weather_observation_url()

        retries = Retry(
            total=number_of_retries, read=number_of_retries, connect=number_of_retries,
            backoff_factor=1, allowed_methods=False,
            status_forcelist=[409] + list(range(500, 600))
        )
        self._session = requests.session()
        self._session.mount('https://', HTTPAdapter(max_retries=retries))

    def get_daily(
        self,
        longitude: float,
        latitude: float,
        start_date: datetime.date,
        end_date: datetime.date,
        timezone: str,
        features: List = None
    ) -> List[Dict]:
        """Get weather daily data.

            Args:
                longitude: longitude for which weather data is needed
                latitude: latitude for which weather data is needed
                start_date: weather data will be provided from this time point
                end_date: weather data will be provided up to this time point
                timezone: the timezone for which the data should be fetched, basically it should represent the
                           timezone of the lat and lng
                features: list of features to return. If None all features in response are returned.
            Returns:
                list with dicts with features for each day.
        """
        start_datetime = datetime.datetime(start_date.year, start_date.month, start_date.day)
        end_datetime = datetime.datetime(end_date.year, end_date.month, end_date.day)

        start_timestamp, end_timestamp = self._get_start_end_timestamp(timezone, start_datetime, end_datetime)
        response = self._make_request(self.DAILY_REQUEST_KEY, longitude, latitude, start_timestamp, end_timestamp)
        return self._parse_response(self.DAY_KEY, response=response, features=features)

    def get_hourly(
        self,
        longitude: float,
        latitude: float,
        start_datetime: datetime,
        end_datetime: datetime,
        timezone: str,
        features: List = None
    ) -> List[Dict]:
        """Get weather hourly data.

            Args:
                longitude: longitude for which weather data is needed
                latitude: latitude for which weather data is needed
                start_datetime: weather data will be provided from this time point
                end_datetime: weather data will be provided up to this time point
                timezone: the timezone for which the data should be fetched, basically it should represent the
                          timezone of the lat and lng
                features: list of features to return. If None all features in response are returned.
            Returns:
                list with dicts with features for each hour
        """
        start_timestamp, end_timestamp = self._get_start_end_timestamp(timezone, start_datetime, end_datetime)
        response = self._make_request(self.HOURLY_REQUEST_KEY, longitude, latitude, start_timestamp, end_timestamp)
        return self._parse_response(self.HOUR_KEY, response=response, features=features)

    def features_list(self, daily: bool) -> List:
        """Get possible features list for weather api"""
        response = self._api.make_request(
            self._api.session.get,
            self._get_weather_feature_url(),
            headers={'Content-Type': 'application/json'}
        )
        if daily:
            request_key = self.DAILY_REQUEST_KEY
            added_key = self.DAY_KEY
        else:
            request_key = self.HOURLY_REQUEST_KEY
            added_key = self.HOUR_KEY

        scheme = response['paths'][f'/weather-observation/{request_key}']['get']['responses']['200']
        property_key = scheme['content']['application/json']['schema']['items']['$ref'].split('/')[-1]
        features = list(response['components']['schemas'][property_key]['properties'].keys())
        features.append(added_key)

        if 'id' in features:
            features.remove('id')

        return features

    def _get_weather_observation_url(self) -> str:
        """Returns url with which weather data is requested."""
        observation_request_fmt = (
            '/weather-observation/{granularity}?'
            'longitude={longitude}&latitude={latitude}&'
            'dateFrom={start_timestamp}&dateTo={end_timestamp}'
        )
        return self._get_weather_url() + observation_request_fmt

    def _get_weather_feature_url(self) -> str:
        return self._get_weather_url() + '/swagger/swagger.json'

    def _get_weather_url(self) -> str:
        """Get weather base Url"""
        return f"{'/'.join(self._api.api_url.split('/')[:-3])}/weather"

    @staticmethod
    def _get_start_end_timestamp(
        timezone: str,
        start_datetime: datetime.datetime,
        end_datetime: datetime.datetime
    ) -> Tuple[int, int]:
        """Returns the start and end timestamps in a given timezone."""
        if start_datetime.tzinfo is not None or end_datetime.tzinfo is not None:
            raise ValueError("start_datetime and end_datetime must be native, aka without timezone info")

        time_zone_obj = pytz.timezone(timezone)
        start_timestamp = int(time_zone_obj.localize(start_datetime).timestamp() * 1000)
        end_timestamp = int(time_zone_obj.localize(end_datetime).timestamp() * 1000)
        return start_timestamp, end_timestamp

    @staticmethod
    def _parse_response(key: str, response: List, features: Optional[List[str]] = None) -> List[Dict]:
        """Parses the response."""
        result = []
        for item in response:
            item_timestamp = int(item['validTimeLocal'])
            item_result = {f: item.get(f) for f in features} if features else item
            item_result[key] = item_timestamp
            result.append(item_result)
        return result

    def _make_request(
        self,
        granularity: str,
        longitude: float,
        latitude: float,
        start_timestamp: int,
        end_timestamp: int
    ) -> List[Dict]:
        """Get weather data.

            Args:
                granularity: either historicalHourly or historicalDaily
                longitude: longitude for which weather data is needed
                latitude: latitude for which weather data is needed
                start_timestamp: get weather data from this timestamp
                end_timestamp: get weather data up to this timestamp
            Returns:
                list of dicts with features key/values as values.
        """
        return self._api.make_request(
            self._session.get,
            self._weather_url_fmt.format(
                granularity=granularity,
                longitude=longitude,
                latitude=latitude,
                start_timestamp=start_timestamp,
                end_timestamp=end_timestamp
            ),
            headers=self._api.headers['processing']
        )



"""
This library provides an interface to the REST API exposed by Ag-MRI.

This library acts as a thin authentication library for the Requests library.
When performing any requests that library documentation should be referenced.

Admin credentials are pulled from the ~/.agmri.cfg file. This file is in the
.ini file format, which is documented here: https://docs.python.org/3.5/library/configparser.html
It should contain a section for each environment with an admin_username and
admin_password value within it.

Example:
```ini
[release]
admin_username = EMAIL
admin_password = PASSWORD

[prod]
admin_username = EMAIL
admin_password = PASSWORD
```

The section headers should match the value that is specified when
initializing the Api() class. An example:

```python
from commons.agmri_api import Api

api = Api("prod")  # This will work with the above cfg file
api = Api("preprod")  # So will this
api = Api("testing")  # This will fail because it's missing from the cfg file
```

"""

import configparser
import logging
import os
from string import Template
from typing import Optional, Tuple, Union, Any, List
from urllib.parse import urlparse

import deprecation
import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError
from urllib3.util.retry import Retry

import intelinair_utils.token_store as token_stores
from intelinair_utils.api import Api
from intelinair_utils.api_utils import validate_date
from intelinair_utils.ssm_utils import get_parameter_from_ssm
from intelinair_utils.token_store import TokenPair

REQUEST_TIMEOUT = 180
RETRIES = 7

logger = logging.getLogger(__name__)


class AgmriApi(Api):
    """
    Wrapper class for authenticated queries against Ag-MRI.

    After initialization, authentication has been completed and requests
    through this object will be automatically authenticated. Only GET,
    POST and PATCH are supported right now.

    URL-resolution precedence (backward compatible):

      1. `api_url` in config → used directly.
      2. `url_template` in config → formatted with {environment}.
      3. Legacy logic:
         • prod  → hard-coded PROD URL
         • other → legacy URL_TEMPLATE

      If both `api_url` and `url_template` are present, `api_url` wins and
      a single warning is logged.
      """

    MAIN_PROD_URL = "https://api.ag-mri.intelinair.com/admin/api/"
    URL_TEMPLATE = "https://api.{}.intelinair.dev/admin/api/"

    def __init__(self, environment: str, config_path: str = None, version_string: str = 'v2/'):
        # version_string == '' corresponds to the first version of API
        if not version_string.endswith('/'):
            raise ValueError(f"all calls (other than login) should use the version string: {version_string}")

        self.environment = environment
        self._config_path = config_path

        self.tokens: TokenPair | None = None
        self._config: configparser.ConfigParser | None = None
        self._token_store = None

        self.api_url = self._resolve_api_url()
        self.api_version_url = self.api_url + version_string
        self.url_template = f"{self.api_version_url}{{}}"

        self.session = requests.Session()
        retries = Retry(
            total=10, read=10, connect=10, backoff_factor=1,
            allowed_methods=False, status_forcelist=[409] + list(range(500, 600))
        )
        self.session.mount('https://', HTTPAdapter(max_retries=retries))

        self.headers = {
            'processing': {'X-Processing': 'true'},
            'graphql': {'Content-Type': 'application/graphql'},
            'refresh': {'Content-Type': 'application/x-www-form-urlencoded'}
        }

    @property
    def config(self) -> configparser.ConfigParser:
        if self._config is None:
            self._config = configparser.ConfigParser()
            self._config['DEFAULT']['token_store'] = 'FileStore'

            if self._config_path is not None:
                parsed_path = urlparse(self._config_path)
                if parsed_path.scheme == "file":
                    self._config.read(parsed_path.path)
                elif parsed_path.scheme == "ssm":
                    config_content = get_parameter_from_ssm(parsed_path.path)
                    self._config.read_string(config_content)
                else:
                    raise Exception("Not supported configuration format. Please use file:// or ssm://")
            else:
                # checking the env variable for config content
                if "AGMRI_API_CONFIG" in os.environ:
                    self._config.read_string(os.environ['AGMRI_API_CONFIG'])
                else:
                    self._config.read(os.path.expanduser("~/.agmri.cfg"))

            if not self._config.has_section(self.environment):
                raise Exception("Missing settings for {} environment in".format(self.environment))
        return self._config

        if not cfg.has_section(self.environment):
            raise KeyError(
                f"Config missing section [{self.environment}]"
            )

        self._config = cfg
        return cfg

    def _resolve_api_url(self) -> str:
        """
        Resolve the base API URL for the chosen environment.

        Precedence:
          1. [env].api_url       – exact value
          2. [env].url_template  – formatted with {environment} or {env},
                                   e.g. https://api-{environment}.newdomain.com/admin/api/
                                        https://api-{env}.newdomain.com/admin/api/
          3. Legacy paths        – prod constant or URL_TEMPLATE

        Always returns a URL ending with exactly one “/”.
        """

        cfg = self.config
        def _with_slash(url: str) -> str:
            """Ensure one (and only one) trailing slash."""
            url = url.strip()
            return url if url.endswith("/") else url + "/"

        def _safe_get(option: str) -> str:
            """Return stripped option value or '' if missing / error / not str."""
            try:
                val = cfg.get(self.environment, option)
            except Exception:
                return ""
            return val.strip() if isinstance(val, str) else ""

        api_raw = _safe_get("api_url")
        tpl_raw = _safe_get("url_template")

        if api_raw:
            if tpl_raw:
                logger.warning(
                    "Both 'api_url' and 'url_template' found for '%s'; using 'api_url'.",
                    self.environment,
                )
            final = _with_slash(api_raw)
            logger.info("API URL via api_url: %s", final)
            return final

        if tpl_raw:
            try:
                rendered = tpl_raw.format(environment=self.environment, env=self.environment)
            except Exception as exc:
                logger.error(
                    "Invalid url_template for '%s' (%s): %s",
                    self.environment, tpl_raw, exc
                )
                raise
            final = _with_slash(rendered)
            logger.info("API URL via url_template: %s", final)
            return final

        if self.environment == "prod":
            final = _with_slash(self.MAIN_PROD_URL)
            logger.info("API URL via legacy PROD constant: %s", final)
            return final

        try:
            final = _with_slash(self.URL_TEMPLATE.format(self.environment))
        except Exception as exc:
            logger.error("Failed to resolve legacy URL for '%s': %s", self.environment, exc)
            raise
        logger.info("API URL via legacy template: %s", final)
        return final

    @property
    def user(self) -> dict:
        return {
            'username': self.config.get(self.environment, 'admin_username'),
            'password': self.config.get(self.environment, 'admin_password')
        }

    @property
    def token_store(self) -> Union[token_stores.SSMStore, token_stores.FileStore]:
        if self._token_store is None:
            driver = self.config.get(self.environment, 'token_store')
            self._token_store = getattr(token_stores, driver)(self.environment, self.user['username'])
        return self._token_store

    def login(self) -> None:
        # login URL does not use version url. it's always /api/login
        logger.info(f'Logging in to {self.api_url} with username={self.user["username"]}')
        auth_response = self.make_request(
            self.session.post,
            self.api_url + "login",
            json=self.user,
            ignore_json=False,
            attempt_authentication=False
        )
        if auth_response is not None:
            self.tokens = TokenPair(
                access_token=auth_response['access_token'],
                refresh_token=auth_response['refresh_token']
            )
            self.headers['processing']['X-Auth-Token'] = auth_response['access_token']
        else:
            raise Exception("Failed to login into ag-mri")

    def token_is_valid(self) -> bool:
        if self.tokens is None or not self.tokens.access_token:
            return False

        url = os.path.join(self.api_url, "validate")
        try:
            res = self.make_request(
                self.session.post,
                url,
                headers={'X-Auth-Token': self.tokens.access_token},
                ignore_json=True,
                attempt_authentication=False
            )
        except HTTPError as e:
            logger.debug(f'token not valid with exception {e}')
            return False
        else:
            if res.status_code == 200:
                return True
            else:
                return False

    def refresh_token(self) -> None:
        url = self.api_url.replace("api/", "oauth/access_token")
        params = {
            'grant_type': 'refresh_token',
            'refresh_token': self.tokens.refresh_token
        }
        res = self.make_request(
            self.session.post,
            url,
            params=params,
            headers=self.headers['refresh'],
            attempt_authentication=False
        )
        self.tokens = TokenPair(access_token=res['access_token'], refresh_token=res['refresh_token'])
        self.headers['processing']['X-Auth-Token'] = res['access_token']

    def ensure_authenticated(self) -> None:
        """This is the main method that ensures that the api is connected."""
        if self.token_is_valid():
            return

        try:
            self.tokens = self.token_store.get_tokens()
            if self.token_is_valid():
                self.headers['processing']['X-Auth-Token'] = self.tokens.access_token
                return
        except Exception as e:
            logger.debug(f'failed to fetch tokens with exception {e}')

        try:
            self.refresh_token()
        except Exception as e:
            logger.debug(f'failed to refresh tokens with exception {e}')
        else:
            self.token_store.store_tokens(self.tokens)
            return

        self.login()
        self.token_store.store_tokens(self.tokens)

    def make_request(
        self, http_method: Any,
        *args: Any,
        ignore_json: bool = False,
        attempt_authentication: bool = True,
        **kwargs: Any
    ) -> Union[Any, requests.Response]:
        """Makes the Session request.

        Returns:
            the response object if ignore_json is True, otherwise response json decoded object
        """
        response = http_method(*args, **kwargs, timeout=REQUEST_TIMEOUT)
        status_code = response.status_code

        if status_code == 401 and attempt_authentication:
            self.ensure_authenticated()
            kwargs['headers'].update(self.headers['processing'])
            return self.make_request(
                http_method, *args, ignore_json=ignore_json, attempt_authentication=False, **kwargs
            )

        response.raise_for_status()

        if status_code == 204:
            return response

        if ignore_json:
            return response
        else:
            return response.json()

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def get(self, path, params=None, ignore_json=False, **kwargs):
        """Perform a GET request to the specified endpoint"""
        return self.make_request(
            self.session.get,
            self.api_version_url + path.lstrip('/'),
            headers=self.headers['processing'],
            params=params,
            ignore_json=ignore_json,
            **kwargs
        )

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def post(self, path, params=None, ignore_json=False, **kwargs):
        # basically this extends http://docs.python-requests.org/en/latest/api/#requests.request
        # in particular, kwargs can contain files array
        """Perform a POST request to the specified endpoint"""
        return self.make_request(
            self.session.post,
            self.api_version_url + path.lstrip('/'),
            headers=self.headers['processing'],
            params=params,
            ignore_json=ignore_json,
            **kwargs
        )

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def delete(self, path, params=None, ignore_json=True):
        """Perform a POST request to the specified endpoint"""
        return self.make_request(self.session.delete, self.api_version_url + path.lstrip('/'),
                                 headers=self.headers['processing'], params=params, ignore_json=ignore_json)

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def post_json(self, path, json=None, ignore_json=False):
        """Perform a POST request to the specified endpoint"""
        return self.make_request(self.session.post, self.api_version_url + path.lstrip('/'),
                                 headers=self.headers['processing'], json=json, ignore_json=ignore_json)

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def patch_json(self, path, json=None, ignore_json=True):
        """Perform a PATCH request to the specified endpoint"""
        return self.make_request(self.session.patch, self.api_version_url + path.lstrip('/'),
                                 headers=self.headers['processing'], json=json, ignore_json=ignore_json)

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def put_json(self, path, json=None, ignore_json=False):
        # TODO set ignore_json=True in after making sure that json problem is not connected with response code
        """Perform a PUT request to the specified endpoint"""
        return self.make_request(self.session.put, self.api_version_url + path.lstrip('/'),
                                 headers=self.headers['processing'], json=json, ignore_json=ignore_json)



    @property
    def _graphql_host(self) -> str:
        """
        Reuse the same host logic as REST, but strip any path.
        E.g. if api_url → "https://foo.bar/admin/api/v2/", returns "foo.bar"
        """
        parsed = urlparse(self._resolve_api_url())
        return parsed.netloc

    def graphql_request(
        self,
        service: str,
        query: str,
        extra_headers: Optional[dict] = None,
        graphql_content_as_json: bool = False,
    ) -> dict:
        """Perform a GraphQL request to the specified endpoint

        Args:
            service: service name, e.g. 'admin', 'reports'
            query: GraphQL query
            extra_headers: extra headers to be added to the request
            graphql_content_as_json: if True, assumes that the graphql query will have JSON content
        Returns:
            GraphQL response dict
        """
        extra_headers = extra_headers or {}
        if 'X-Auth-Token' not in self.headers['processing']:
            self.ensure_authenticated()

        if graphql_content_as_json:
            graphql_headers = {'Content-Type': 'application/json'}
            graphql_headers.update({k: v for k, v in self.headers['graphql'].items() if k != 'Content-Type'})
        else:
            graphql_headers = self.headers['graphql']

        headers = {**self.headers['processing'], **graphql_headers, **extra_headers}

        host = self._graphql_host
        graphql_url = "https://{}/{}/graphql".format(host, service)

        response = self.make_request(self.session.post, graphql_url, headers=headers, data=query)
        return response

    def fields_report(
        self,
        request_fields: Union[List[str], Tuple[str, ...]],
        extent: Optional[Tuple[int, int, int, int]] = None,
        only_monitored: bool = True,
        enabled: bool = True,
        company_deleted: bool = False,
        prospecting_mode=None,
        only_in_grid: str = None,
        satellite_only: bool = False,
        company_id: int = None,
        provider: str = None,
        flight_date: str = None,
        r_max: int = 10000,
        offset: int = 0
    ):

        query = Template("""{
          fieldsReport(queryCommand: {
            $params
          }, max: $r_max, offset: $offset) {
                totalCount
                results {
                    $results
                }
            }
        }""")

        params = ""
        if extent is not None:
            params += """northEastLatitude: {},
            northEastLongitude: {},
            southWestLatitude: {},
            southWestLongitude: {},""".format(extent[0], extent[1], extent[2], extent[3])
        if only_monitored is True:
            params += """
            onlyMonitored: true,"""
        if enabled is True:
            params += """
            enabled: true,"""
        if company_deleted is False:
            params += """
            companyDeleted: false,"""
        if only_in_grid is True:
            params += """
            onlyInGrid: true,"""
        if satellite_only is True:
            params += """
            satelliteOnly: true,"""
        if prospecting_mode:
            params += """
            prospectingMode: {},""".format(prospecting_mode)
        if company_id:
            params += """
            companyId: {},""".format(company_id)
        if provider is not None:
            params += """
            provider: \"{}\",""".format(provider)
        if flight_date is not None:
            params += """
            flightDate: \"{}\",""".format(flight_date)
        params += """
            sort: \"id\""""

        results = ""
        if 'id' in request_fields:
            results += "id"
        if 'name' in request_fields:
            results += """
                    name"""
        if 'area' in request_fields:
            results += """
                    area"""
        if 'tags' in request_fields:
            results += """
                    tags"""
        if 'description' in request_fields:
            results += """
                    description"""
        if 'grid_id' in request_fields:
            results += """
                    grid{
                        id
                    }"""
        if 'grid_ids' in request_fields:
            results += """
                    gridIds"""
        if 'prospecting_mode' in request_fields:
            results += """
                    farm{
                        division{
                            company{
                                alias{
                                    companyHierarchy{
                                      prospectingMode
                                    }
                                }
                            }
                        }
                    }"""
        if 'latestFlight' in request_fields:
            results += """
                    latestFlight{
                        code
                    }"""
        if 'geometry' in request_fields:
            results += """
                    geoData{
                        geometry
                    }"""

        query = query.substitute(params=params, results=results, r_max=r_max, offset=offset)
        data = self.graphql_request(service="reports", query=query)
        return data

    def get_field_info(self, field_id: int, request_fields: List[str], service: str = 'admin'):
        """get field info

        Args:
            field_id: field id
            request_fields: fields to request available fields are
                id, name, latestFlight, fieldSeason, cropTypes, plantingDate, tags, area, token, plantedSeeds, flights
            service: service to make a call. Defaults to 'admin'.

        Returns:
            dict: field info
        """
        query = Template(
            """{
              field(id: $field_id) {
                $params
                } 
            }"""
        )
        params = ""

        if 'id' in request_fields:
            params += """
            id"""
        if 'name' in request_fields:
            params += """
            name"""
        if 'latestFlight' in request_fields:
            params += """
            latestFlight{
                id
                date
            }"""
        if 'fieldSeason' in request_fields:
            params += """
            fieldSeason{
                id
                geoData{
                    id
                    geometry
                }
            }"""
        if 'cropTypes' in request_fields:
            params += """
            cropTypes {
                id
                name
            }"""
        if 'plantingDate' in request_fields:
            params += """
            plantingDate"""
        if 'tags' in request_fields:
            params += """
            tags"""
        if 'area' in request_fields:
            params += """
            area"""
        if 'token' in request_fields:
            params += """
            token"""
        if 'plantedSeeds' in request_fields:
            params += """
            plantedSeeds"""
        if 'flights' in request_fields:
            params += """
            flights {
                id,
                date,
                code
                season {
                    id
                }
                provider
                reviewStatus
            }"""
        query = query.substitute(params=params, field_id=field_id)
        data = self.graphql_request(service=service, query=query)
        return data.get('data', {}).get('field', {})

    def get_attachment_types(self) -> dict:
        types = self.get("fieldattachmenttypes?max=100")
        return {z['keyName']: z['id'] for z in types['fieldAttachmentTypes']}

    # GET requests
    def get_flight(self, flight_code: str) -> dict:
        """Getting flight object."""
        url = self.url_template.format(f"flights/{flight_code}")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_flight_features(self, flight_code: str) -> dict:
        """Getting flight features."""
        url = self.url_template.format(f"flights/{flight_code}/features")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_imagery(self, imagery_code: str) -> dict:
        """Getting imagery data."""
        url = self.url_template.format(f"imagery/{imagery_code}")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_seasons(self) -> dict:
        """
        Getting season info.
        Returns:
            list of seasons info wrapped in dict.
        """
        url = self.url_template.format("season/")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_field(self, field_id: Union[int, str]) -> dict:
        """Getting field info for a given field_id or field_token."""
        url = self.url_template.format(f"fields/{field_id}")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_field_products(self, field_id: Union[int, str], season_id: Union[int, str]) -> dict:
        """Getting field products for a given field_id and season_id."""
        url = self.url_template.format(f"fieldInfo/{field_id}/{season_id}")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_flight_processing_data(self, flight_code: str) -> dict:
        """ Getting processing data for the flight."""
        url = self.url_template.format(f"flights/{flight_code}/processing")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_equipment_flight_data(self, flight_code: str) -> dict:
        """Getting equipment flight data."""
        url = self.url_template.format(f"flights/{flight_code}/equipmentData")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_uploaded_file(self, file_id: Union[int, str]) -> dict:
        """Getting uploaded file by ID."""
        url = self.url_template.format(f"uploadFile/{file_id}/")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_uploaded_files(self, params: dict) -> dict:
        """
        Getting uploaded files based on filters in params.
        Args:
            params: http request params as key values.
        """
        url = self.url_template.format("uploadFile/")
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def get_field_attachments(self, field_id: Union[int, str]) -> list:
        """Getting field attachments list."""
        url = self.url_template.format(f"fields/{field_id}/attachments")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_field_processing_data(self, field_id: Union[int, str]) -> list:
        """Get processing data for the field."""
        url = self.url_template.format(f"fields/{field_id}/processing")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_field_all_boundaries(self, field_id: Union[int, str], ignore_json: bool = True) -> dict:
        """Getting field all boundaries."""
        url = self.url_template.format(f"fields/{field_id}/allboundaries")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'], ignore_json=ignore_json)
        return response

    def get_previous_flight(self, flight_code: str) -> dict:
        """Getting previous flight of the given one."""
        url = self.url_template.format(f"flights/{flight_code}/previous")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_field_tilesets(self, field_id: Union[int, str]) -> dict:
        """Getting list if tilesets for a given field."""
        url = self.url_template.format(f"fields/{field_id}/tilesets")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def refresh_field_report(self) -> None:
        """Refresh field report."""
        url = self.url_template.format("fields/refreshFieldReport")
        self.make_request(self.session.get, url, headers=self.headers['processing'])

    def get_flight_alerts(self, flight_code: str, alert_name: str) -> dict:
        """
        Getting flight alerts for the given alert name and flight code.
        Args:
            flight_code:
            alert_name:
        """
        url = self.url_template.format(f"flights/{flight_code}/alerts?kindKeys={alert_name}")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_imageries(self, filters: dict) -> dict:
        """
        Getting imageries filtered by given constraints in filters.
        Args:
            filters: key values based on which output will be generated. e.g.
            {
                "startDate": 2022-05-01,
                "endDate": 2022-05-31,
                "seasonId": 71,
                "provider": "Sentinel2"
            }
        """
        url = self.url_template.format("imagery/")
        response = self.make_request(self.session.get, url, params=filters, headers=self.headers['processing'])
        return response

    @deprecation.deprecated(details='Use get_growing_degree_days_by_field instead.')
    def get_growing_degree_days(self, flight_code: str, date_from: str, date_to: str) -> dict:
        """
        Get GDD for the flight.
        Args:
            flight_code:
            date_from:
            date_to:
        """
        validate_date(date_from)
        validate_date(date_to)
        params = {
            'flightCode': flight_code,
            'dateFrom': date_from,
            'dateTo': date_to
        }
        url = self.url_template.format("flights/gddForFlight")
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def get_growing_degree_days_by_field(self, field_id: int, date_from: Optional[str], date_to: str) -> dict:
        """Get GDD for a given field and date range."""
        if date_from:
            validate_date(date_from)
        validate_date(date_to)
        params = {
            'fieldId': field_id,
            'dateFrom': date_from,
            'dateTo': date_to
        }
        url = self.url_template.format("fields/gdd")
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def get_last_frost_date(self, field_id: Union[int, str], season_id: int, date_to: str) -> dict:
        """
        Get last frost date for the field.
        Args:
            field_id: the field id or token
            season_id: the season id
            date_to: the date before which the last frost will be returned
        """
        validate_date(date_to)
        url = self.url_template.format(f'fields/{field_id}/lastFrost')
        params = {
            'seasonId': season_id,
            'dateTo': date_to
        }
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def get_rollback_gdd_date(self, field_id: Union[int, str], season_id: int, date_from: str, rollback=100) -> dict:
        """
        Get rollback gdd date for the field.
        Args:
            field_id: the field id or token
            season_id: season id
            date_from: the date from which GDD should be rolled back
            rollback: the amount of GDD should be rolled back
        """
        validate_date(date_from)
        url = self.url_template.format(f'fields/{field_id}/rollbackGdd')
        params = {
            'seasonId': season_id,
            'dateFrom': date_from,
            'delta': rollback
        }
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def get_field_planting_date_and_crop_types(
        self,
        field_id: Union[str, int],
        season_id: int
    ) -> Tuple[Optional[str], list]:
        """Gets the available planting date and crop types from the api

        Args:
            field_id: the field id
            season_id: the season id

        Returns:
            planting datetime and list of crop types
        """
        query = f"""{{
                  listFieldSeasons(queryCommand: {{seasonId: {season_id}, fieldId: {field_id}}}) {{
                    results {{
                      cropTypeKeys
                      plantingDate
                    }}
                  }}
                }}"""
        query_results = self.graphql_request(service='admin', query=query)
        if not query_results['data']['listFieldSeasons']['results']:
            return None, []
        [field_season] = query_results['data']['listFieldSeasons']['results']
        if field_season.get('cropTypeKeys') is not None:
            crop_types = [ct.split('.')[-1].lower() for ct in field_season['cropTypeKeys']]
        else:
            crop_types = []
        planting_date = field_season.get('plantingDate')
        return planting_date, crop_types

    def get_field_attachment_types(self, max_count: int = 100) -> list:
        """
        Getting available attachment types for the field.
        Args:
            max_count: Maximum count of types to return.
        """
        url = self.url_template.format("fieldattachmenttypes")
        params = {
            'max': max_count
        }
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def check_field_weather_data_existence(self, field_id: Union[int, str], start_date: str, end_date: str) -> dict:
        """Checking given field weather data download completion status between given date range."""
        validate_date(start_date)
        validate_date(end_date)
        url = self.url_template.format("weather/verifyConsistency")
        params = {
            'fieldId': field_id,
            'startDate': start_date,
            'endDate': end_date
        }
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def get_company(self, company_id: Union[int, str]) -> dict:
        """Getting company info from given company id."""
        url = self.url_template.format(f"companies/{company_id}")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_flight_tilesets(self, flight_code: str) -> dict:
        """Getting flight tilesets."""
        url = self.url_template.format(f"flights/{flight_code}/tilesets")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_problem_alert_default_tileset(self, flight_code: str, badge_id: Union[int, str]) -> dict:
        """
        Getting default tileset for problem alert.
        Args:
            flight_code: the flight code for which the alerts should be returned
            badge_id: problem alert badge ID
        """
        params = {
            'badgeId': badge_id
        }
        url = self.url_template.format(f"flights/{flight_code}/defaultTileset")
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    # POST requests
    def imageries_report(self, filters: dict) -> dict:
        """
        Getting reported imagery list based on filters.
        Args:
            filters: key values based on which output will be generated. e.g.
            {
                "startDate": 2022-05-01,
                "endDate": 2022-05-31
            }
        """
        url = self.url_template.format("reports/imagery/")
        response = self.make_request(self.session.post, url, params=filters, headers=self.headers['processing'])
        return response

    def upload_complete_flight(self, flight_code: str) -> None:
        """Setting upload Complete to the flight."""
        url = self.url_template.format(f"flights/{flight_code}/uploadComplete")
        self.make_request(self.session.post, url, headers=self.headers['processing'])

    def get_field_equipment_flight_data(self, field_id: Union[int, str], season_id: int) -> dict:
        """Getting list of equipment flights data for the field."""
        url = self.url_template.format("flights/equipmentFlightsData")
        params = {
            "fieldId": field_id,
            "seasonId": season_id
        }
        response = self.make_request(self.session.post, url, params=params, headers=self.headers['processing'])
        return response

    def upload_field_attachment(self, field_id: Union[int, str], attachment_id: Union[int, str], attachment_path: str):
        """
        Uploading field attachment file.
        Attachment is done through binary stream.
        Args:
            field_id: the field id for which attachment should be uploaded
            attachment_id: the id of the attachment
            attachment_path: the path of tha attachment to be uploaded
        """
        url = self.url_template.format(f"fields/{field_id}/attachments/{attachment_id}/upload")
        with open(attachment_path, 'rb') as attachment_handle:
            files = {
                "attachmentFile": (os.path.basename(attachment_path), attachment_handle, 'application/octet-stream')
            }
            self.make_request(self.session.post, url, files=files, headers=self.headers['processing'])

    def rerun_flight(self, flight_code: str) -> None:
        """Rerunning given flight processing."""
        url = self.url_template.format(f"flights/{flight_code}/rerun")
        self.make_request(self.session.post, url, headers=self.headers['processing'])

    # DELETE requests
    def delete_flight(self, flight_code: str) -> None:
        """Delete flight by code."""
        url = self.url_template.format(f"flights/{flight_code}")
        self.make_request(self.session.delete, url, headers=self.headers['processing'], ignore_json=True)

    def delete_field_attachment(self, field_id: Union[int, str], attachment_id: Union[int, str]) -> None:
        """Delete field attachment file by given field and file id."""
        url = self.url_template.format(f"fields/{field_id}/attachments/{attachment_id}")
        self.make_request(self.session.delete, url, headers=self.headers['processing'], ignore_json=True)

    def delete_field_attachments_by_file_type(self, field_id: int, flight_id: Optional[int], file_type_id: int) -> None:
        """Deletes all the field attachments matching the given criteria

        Args:
            field_id: the id of the field to delete attachments for
            flight_id: if provided will only delete attachments that have this flight_id
            file_type_id: only deletes attachments with this file_type_id
        """
        query = f"""query {{
        field(id: {field_id}) {{ id fieldAttachments{{ id version fileType {{ id }} flight {{ id }}}}}}}}
        """
        result = self.graphql_request(service='admin', query=query)
        attachments = result['data']['field']['fieldAttachments']
        attachments = filter(lambda x: x['fileType']['id'] == file_type_id, attachments)
        if flight_id:
            attachments = filter(lambda x: x['flight'] is not None, attachments)
            attachments = filter(lambda x: x['flight']['id'] == flight_id, attachments)

        attachments = list(attachments)
        logger.info(f'Found {len(attachments)} existing veg row field attachments for this flight, will delete them')
        for att in attachments:
            logger.info(f'Deleting Field Attachment with id={att["id"]}')
            self.delete_field_attachment(field_id=field_id, attachment_id=att['id'])

    def delete_field_tileset(self, field_id: Union[int, str], tileset_id: Union[int, str]) -> None:
        """Delete field tileset by id."""
        url = self.url_template.format(f"fields/{field_id}/tilesets/{tileset_id}")
        self.make_request(self.session.delete, url, headers=self.headers['processing'], ignore_json=True)

    def delete_flight_tileset(self, flight_id: Union[int, str], tileset_id: Union[int, str]) -> None:
        """Delete flight tileset by id."""
        url = self.url_template.format(f"flights/{flight_id}/tilesets/{tileset_id}")
        self.make_request(self.session.delete, url, headers=self.headers['processing'], ignore_json=True)

    # POST_JSON requests
    def get_flights_annotations(self, filters: dict, params=None) -> dict:
        """
        Getting flight annotations report based on input params in filter.
        Args:
            filters: key values based on which output will be generated. e.g.
            {
                "alertKindId": 29,
                "annotationStatus": "QA_COMPLETE"
            }
            params: http request params as key values.
        """
        url = self.url_template.format("reports/toafa")
        response = self.make_request(self.session.post, url, filters, params=params, headers=self.headers['processing'])
        return response

    def add_field_tag(self, field_id: Union[int, str], tag: str, key_name: str = 'dashboard') -> None:
        """
        Add tag to field object.
        Args:
            field_id: the field id or token
            tag: Tag name to add.
            key_name: tag keyname (mainly dashboard and system).
        """
        url = self.url_template.format(f"fields/{field_id}/addTag")
        params = {
            "optionName": tag,
            "keyName": key_name
        }
        self.make_request(self.session.post, url, json=params, headers=self.headers['processing'])

    def remove_field_tag(self, field_id: Union[int, str], tag: str, key_name: str = 'dashboard') -> None:
        """
        Remove given tag from the field object.
        Args:
            field_id: the field id or token
            tag: tag name
            key_name: tag keyname (mainly dashboard and system).
        """
        url = self.url_template.format(f"fields/{field_id}/removeTag")
        params = {
            "optionName": tag,
            "keyName": key_name
        }
        self.make_request(self.session.post, url, json=params, headers=self.headers['processing'])

    def add_flight_tag(self, flight_code: str, tag: str, key_name: str = 'dashboard') -> None:
        """
        Add tag to the flight object.
        Args:
            flight_code: flight code
            tag: tag name to add
            key_name: tag keyname (mainly dashboard and system).
        """
        url = self.url_template.format(f"flights/{flight_code}/addTag")
        params = {
            "optionName": tag,
            "keyName": key_name
        }
        self.make_request(self.session.post, url, json=params, headers=self.headers['processing'])

    def remove_flight_tag(self, flight_code: str, tag: str, key_name: str = 'dashboard') -> None:
        """
        Remove given tag from the flight object.
        Args:
            flight_code: flight code
            tag: tag name
            key_name: tag keyname (mainly dashboard and system).
        """
        url = self.url_template.format(f"flights/{flight_code}/removeTag")
        params = {
            "optionName": tag,
            "keyName": key_name
        }
        self.make_request(self.session.post, url, json=params, headers=self.headers['processing'])

    def get_field_flight(self, field_id: Union[int, str], filters: dict) -> dict:
        """
        Getting flight for a given field based on input filters.
        Args:
            field_id:
            filters: key values based on which output will be generated. e.g.
            {
                'provider': 'Aeroptic',
                'date': 2022-05-01,
            }
        """
        url = self.url_template.format(f"fields/{field_id}/flight")
        response = self.make_request(self.session.post, url, json=filters, headers=self.headers['processing'])
        return response

    def flights_report(self, filters: dict) -> dict:
        """
        Generates report of the flights based on a given filters.
        Args:
            filters: key values based on which output will be generated. e.g.
            {
                'fieldId': 12345678,
                'providers': ['Aeroptic'],
                'startDate': 2022-05-01,
                'endDate': 2022-05-30
            }
        """
        url = self.url_template.format("reports/toaf")
        response = self.make_request(self.session.post, url, json=filters, headers=self.headers['processing'])
        return response

    def create_imagery(self, payload: dict) -> dict:
        """
        Creates imagery object with values passed through payload.
        Args:
            payload: properties of imagery object to create.
        Returns:
            Created imagery object.
        """
        url = self.url_template.format("imagery/")
        response = self.make_request(self.session.post, url, json=payload, headers=self.headers['processing'])
        return response

    def delete_imagery(self, imagery_code: str) -> None:
        """Deletes imagery object with values passed through payload."""
        url = self.url_template.format(f"imagery/{imagery_code}")
        response = self.make_request(self.session.delete, url, headers=self.headers['processing'], ignore_json=True)
        response.raise_for_status()

    def create_flight(self, payload: dict) -> dict:
        """
        Creates flight object with values passed through payload.
        Args:
            payload: properties of flight object to create.
        Returns:
            Created flight object.
        """
        url = self.url_template.format("flights/")
        response = self.make_request(self.session.post, url, json=payload, headers=self.headers['processing'])
        return response

    def create_imagery_tile_layer(self, payload: dict) -> None:
        """
        Creates imagery layer.
        Args:
            payload: properties of tile layer to create.
        """
        url = self.url_template.format("imagery/tiles/layers")
        self.make_request(self.session.post, url, json=payload, headers=self.headers['processing'])

    def post_field_metadata(self, field_id: Union[int, str], metadata: dict) -> requests.Response:
        """Create field metadata."""
        url = self.url_template.format(f"fields/{field_id}/metadata/")
        response = self.make_request(self.session.post, url, json=metadata, headers=self.headers['processing'],
                                     ignore_json=True)
        return response

    def toggle_flight_tileset(self, flight_code: str, tileset_name: str, is_hidden: bool) -> None:
        """
        Changing hidden attribute of the flight tileset.
        Args:
            flight_code: on which change should be performed.
            tileset_name: tileset which will be changed.
            is_hidden: value to be set.
        """
        url = self.url_template.format(f"flights/{flight_code}/toggleTilesets/")
        payload = {
            'tilesets': [
                {
                    'name': tileset_name,
                    'hidden': is_hidden
                }
            ]
        }
        self.make_request(self.session.post, url, json=payload, headers=self.headers['processing'],
                          ignore_json=True)

    def reject_flight(self, flight_code: str, comment: str) -> None:
        """Automatically reject the flight with a given comment."""
        url = self.url_template.format(f"flights/{flight_code}/flightComments/")
        payload = {
            "reviewer": "Automatic",
            "flightCode": flight_code,
            "subject": "Automatic rejection",
            "comments": comment
        }
        self.make_request(self.session.post, url, json=payload, headers=self.headers['processing'])

    def create_field_attachment(self, field_id: Union[int, str], attachment_data: dict) -> dict:
        """Create field attachment with given data."""
        url = self.url_template.format(f'/fields/{field_id}/attachments')
        response = self.make_request(self.session.post, url, json=attachment_data, headers=self.headers['processing'])
        return response

    def create_field_tilesets(self, field_id: Union[int, str], tilesets_data: dict) -> None:
        """Creates tilesets from data for the given field."""
        url = self.url_template.format(f'/fields/{field_id}/tilesets')
        self.make_request(self.session.post, url, json=tilesets_data, headers=self.headers['processing'])

    def create_flight_tilesets(self, flight_code: str, tilesets_data: dict) -> None:
        """Creates tilesets from data for the given flight."""
        url = self.url_template.format(f'/flights/{flight_code}/tilesets')
        self.make_request(self.session.post, url, json=tilesets_data, headers=self.headers['processing'])

    # PATCH_JSON requests
    def update_imagery(self, imagery_code: str, imagery_data: dict) -> None:
        """
        Update imagery object properties.
        Args:
            imagery_code: code of imagery to be changed.
            imagery_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"imagery/{imagery_code}/")
        self.make_request(self.session.patch, url, json=imagery_data, headers=self.headers['processing'],
                          ignore_json=True)

    def update_field(self, field_id: Union[int, str], field_data: dict) -> None:
        """
        Update field object properties.
        Args:
            field_id: id of field to be changed.
            field_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"fields/{field_id}/")
        self.make_request(self.session.patch, url, json=field_data, headers=self.headers['processing'],
                          ignore_json=True)

    def update_field_attachment(
        self,
        field_id: Union[int, str],
        attachment_id: Union[int, str],
        attachment_data: dict
    ) -> None:
        """
        Update field attachment file
        Args:
            field_id: the field id or token
            attachment_id: id of a file to be changed
            attachment_data: key-values of properties that will be changed
        """
        url = self.url_template.format(f"fields/{field_id}/attachments/{attachment_id}/")
        self.make_request(self.session.patch, url, json=attachment_data, headers=self.headers['processing'],
                          ignore_json=True)

    def update_flight(self, flight_code: str, flight_data: dict) -> None:
        """
        Update flight object.
        Args:
            flight_code: code of flight to change
            flight_data: key-values of properties that will be changed
        """
        url = self.url_template.format(f"flights/{flight_code}/")
        self.make_request(self.session.patch, url, json=flight_data, headers=self.headers['processing'],
                          ignore_json=True)

    def update_flight_features(self, flight_code: str, features: dict) -> None:
        """
        Update flight features.
        Args:
            flight_code: code of flight to change.
            features: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"flights/{flight_code}/features")
        self.make_request(self.session.post, url, json=features, headers=self.headers['processing'],
                          ignore_json=False)

    def update_flight_tilesets(
        self,
        flight_id: Union[int, str],
        tileset_id: Union[int, str],
        tilesets_data: dict
    ) -> None:
        """Update given flight tilesets."""
        url = self.url_template.format(f"flights/{flight_id}/tilesets/{tileset_id}")
        self.make_request(self.session.patch, url, json=tilesets_data, headers=self.headers['processing'],
                          ignore_json=True)

    # PUT_JSON requests
    def update_uploaded_file(self, file_id: Union[int, str], file_data: dict) -> None:
        """
        Update uploaded file.
        Args:
            file_id: id of file to change
            file_data: key-values of properties that will be changed
        """
        url = self.url_template.format(f"uploadFile/{file_id}/")
        self.make_request(self.session.put, url, json=file_data, headers=self.headers['processing'])

    def update_flight_alert(self, flight_code: str, alert_id: Union[int, str], alert_data: dict) -> None:
        """
        Update alert of the flight.
        Args:
            flight_code: code of a flight to change.
            alert_id: id of an alert to change.
            alert_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"flights/{flight_code}/alerts/{alert_id}/")
        self.make_request(self.session.put, url, json=alert_data, headers=self.headers['processing'])

    def add_field_list_to_file(self, file_id: Union[int, str], field_list: list) -> None:
        """
        add field list to the uploaded file.
        Args:
            file_id: uploaded file id
            field_list: list of fields to be added
        """
        url = self.url_template.format(f"uploadFile/{file_id}/addToFieldList")
        payload = {
            "fields": field_list
        }
        self.make_request(self.session.put, url, json=payload, headers=self.headers['processing'])

    def delete_tileset(self, tileset_id: int) -> None:
        """
        Deletes tileset with the given id.
        Args:
            tileset_id: the id of the tileset to be deleted
        """
        url = self.url_template.format(f"tilesets/{tileset_id}")
        self.make_request(self.session.delete, url, headers=self.headers['processing'], ignore_json=True)



"""Planting service API wrapper class."""

from typing import Dict, List, Optional

from intelinair_utils.api_wrapper import ApiWrapper


class VectorTileServiceApi(ApiWrapper):
    """Planting service API class."""
    VECTOR_TILE_SERVICE_URLS = {
        "bellflower": "https://pi0m191jpc.execute-api.us-east-1.amazonaws.com/api",
        "release": "https://4yboi7re8i.execute-api.us-east-1.amazonaws.com/api",
        "prod": "https://3qs6tciagc.execute-api.us-east-1.amazonaws.com/api",
    }

    S3_CACHE_PATHS = {
        "bellflower": "s3://intelinair-dev-private/bellflower/apicache/vector-tile-service",
        "release": "s3://intelinair-stage-private/release/apicache/vector-tile-service",
        "prod": "s3://intelinair-prod-private/prod/apicache/vector-tile-service",
    }

    def __init__(self, environment: str, config_path: Optional[str] = None):
        super().__init__(environment, self._get_api_urls(), config_path=config_path)

    def _get_api_urls(self):
        """Overriden method for getting planting service API gateway URLs."""
        return self.VECTOR_TILE_SERVICE_URLS

    def get_s3_cache_paths(self):
        """Overriden method for getting planting service S3 cache paths."""
        return self.S3_CACHE_PATHS

    def create_vector_tile(self, data: Dict):
        """
        Method for creating tile geometry along with metadata and properties
        Args:
            data: data is a dictionary of keys
            such as geometry, properties, metadata

        Returns:
            token
        """
        url = self.url_template.format(f"polygon")
        response = self._make_large_body_request(self.session.post, url, json=data)

        return response.json()

    def get_tile_mvt(self, token, z, x, y):
        """
        Methods for transforming hex-geometry to MVT
        Args:
            token: unique identifier for a tile (UUID)
            x, y, z : MVT coordinates

        Returns:
            MVT transformed geometry
        """
        url = self.url_template.format(f"polygon/{token}/{z}/{x}/{y}")
        response = self._make_request(self.session.get, url)

        return response

    def get_tile_geometry(self, token: str):
        """
        Methods that fetches geometry of a tile in geojson
        Args:
            token: unique identifier for a tile (UUID)

        Returns:
            geometry of a vector tile
            properties

        """
        url = self.url_template.format(f"polygon/{token}")
        json_response = self._custom_chunked_get(url)
        return json_response

    def get_tile_metadata(self, token):
        """
        Method that fetches metadata of a vector tile
        Args:
            token: unique identifier for a tile (UUID)

        Returns:
            vector tile metadata

        """
        url = self.url_template.format(f"polygon/{token}/metadata")
        response = self._make_request(self.session.get, url)

        return response.json()

    def create_feature_collection(self, features: List[Dict], metadata: Dict = None) -> Dict:
        """
        Creates a feature collection
        Args:
            features: list of geojson features

        Returns:
            response from the API
        """
        url = self.url_template.format("feature-collections")
        payload = {
            "features": features,
            "meta_data": metadata
        }
        response = self._make_large_body_request(self.session.post, url, json=payload)
        return response.json()

    def get_feature_collection(self, token: str) -> Dict:
        """
        Retrieves a feature collection for a given token
        Args:
            token: unique identifier for a feature collection

        Returns:
            feature collection resource (not to be confused with geojson feature collection format)

        """
        url = self.url_template.format(f"feature-collections/{token}")
        json_response = self._custom_chunked_get(url)

        return json_response



import io
import json
import zipfile
from pathlib import Path
from typing import Dict, List, Union, Optional, Generator

import geojson
import shapely.geometry
from shapely.validation import make_valid

from intelinair_utils.agmri_api import AgmriApi

ANNOTATION_URL_TEMPLATE = 'https://api.{}.intelinair.dev/admin/graphql/index'


class AnnotationsApi:
    """Special API for annotation tool"""

    ANNOTATION_CRS = 4326

    def __init__(self, agmri_env: str, config_path: Optional[str] = None):
        """
        Args:
            agmri_env: the env name
            config_path: the optional path to agmri.cfg, if yoy want to use ssm, use ssm://<ssm param name>
        """
        self.api = AgmriApi(agmri_env, config_path=config_path)

    def list_annotation_types(self, api_limit_size: int = 100):
        offset = 0
        while True:
            query = f"""
            query {{ 
              listTagOptions( 
                max: {api_limit_size}
                offset: {offset} 
                queryCommand: {{
                  domain: "Annotation"
                }}
              ) {{ 
                results {{ 
                  id 
                  name 
                  description 
                  type {{
                    id
                    name
                    keyName
                    domainName
                    __typename
                  }}
                  color   
                  __typename 
                }} 
                totalCount 
                __typename 
              }} 
            }} 
            """

            response = self.api.graphql_request('admin', query=query)

            for tag_option in response['data']['listTagOptions']['results']:
                yield tag_option

            if api_limit_size + offset < response['data']['listTagOptions']['totalCount']:
                offset += api_limit_size
            else:
                break

    def get_annotation_project_by_name(self, name: str):
        query = f"""query {{getProject(name: \"{name}\") {{
              id
              name
              version
              description
              status
              flightsCount
              labels {{
                id
                version
                color
                createdBy
                dateCreated
                name
                description
                type {{
                  id
                  keyName
                  name
                }}
              }}
            }}
        }}
        """
        return self.api.graphql_request('admin', query)['data']['getProject']

    def create_annotation_project(self, name: str, description: str = "", label_ids: List[int] = None):
        """Create annotation project.

        Args:
            name: name of created project
            description: description for the project
            label_ids: labels used in the project

        Returns: integer project id

        """
        api_key = 'projectCreate'
        payload = f"""mutation {{
          {api_key}(project: {{
                        name: "{name}",
                        description : "{description}",
                        labelIds: {str([] if label_ids is None else label_ids)}
                                }}
                        )
                        {{
                            id
                            name
                            description
                            labelIds
                            status
                            errors {{
                                message
                            }}
                        }}
                    }}"""

        response = self._make_request(payload=payload, api_key=api_key)
        return response['data'][api_key]['id']

    def list_annotation_projects(self, api_limit_size=100):
        offset = 0
        while True:
            query = f"""
            query {{ 
              listProjects( 
                max: {api_limit_size} 
                offset: {offset} 
                queryCommand: {{}}
              ) {{ 
                results {{ 
                  id 
                  name 
                  description 
                  labels {{ 
                    id 
                    name 
                    color 
                    type {{ 
                      id 
                      domainName 
                      __typename 
                    }} 
                    __typename 
                  }} 
                  members {{ 
                    id 
                    user {{
                      id 
                      token 
                      firstname 
                      lastname 
                      __typename 
                    }} 
                    __typename 
                  }} 
                  status 
                  configs 
                  __typename 
                }} 
                totalCount 
                __typename 
              }} 
            }} 
            """

            response = self.api.graphql_request('admin', query=query)

            for project in response['data']['listProjects']['results']:
                yield project

            if api_limit_size + offset < response['data']['listProjects']['totalCount']:
                offset += api_limit_size
            else:
                break

    def delete_annotation_project(self, project_id: int):
        query = f"""
        mutation  {{
          projectDelete(id: {project_id}) {{
            success
            error
            __typename
          }}
        }}
        """

        response = self.api.graphql_request('admin', query=query)
        assert response['data']['projectDelete']['success'], 'Failed to delete project'

    def create_annotatable_flight(self, project_id: int, flight_id: int):
        """Create annotatable flight out of ordinary flight.

        Args:
            project_id: project in which to create annotatable flight
            flight_id: ordinary flight id

        Returns:
            id of created annotatable flight

        """
        api_key = 'annotatableFlightCreate'
        payload = f"""mutation {{
                  {api_key}(annotatableFlight: {{
                                project : {{
                                      id : {project_id}
                                    }}
                                flight : {{
                                     id : {flight_id}
                                    }}
                                }}
                  )
                  {{
                    id
                    errors {{
                        message
                    }}
                  }}
                }}"""
        response = self._make_request(payload=payload, api_key=api_key)
        return response['data'][api_key]['id']

    def list_annotatable_flights(self, project_id: int, api_limit_size: int = 100):
        offset = 0
        while True:
            query = f"""
            query {{
              listAnnotatableFlights( 
                max: {api_limit_size}
                offset: {offset} 
                queryCommand: {{
                  projectId: {project_id}
                }}
              ) {{ 
                results {{ 
                  id 
                  assignee {{
                    id
                    user {{
                      id
                      username
                      __typename
                    }}
                    __typename
                  }}
                  status
                  project {{
                    id
                    __typename
                  }}
                  flight {{
                    id
                    code
                    __typename
                  }}
                  notes
                  groupNumber   
                  __typename 
                }} 
                totalCount 
                __typename 
              }} 
            }} 
            """

            response = self.api.graphql_request('admin', query=query)

            for annotatable_flight in response['data']['listAnnotatableFlights']['results']:
                yield annotatable_flight

            if api_limit_size + offset < response['data']['listAnnotatableFlights']['totalCount']:
                offset += api_limit_size
            else:
                break

    def add_annotations(self, annotatable_flight_id: int, polygons: Dict, labels: List[str], color: str = "#FF9A34"):
        """Add annotation to annotatable flight.

        Args:
            annotatable_flight_id: id of flight to add annotation
            polygons: polygons to add as annotation
            labels: string annotation labels
            color: adding custom color if needed

        Returns:
            response from api call

        """
        api_key = "annotationCreate"
        valid_geom = geojson.dumps(polygons).replace('"', '\\"')

        payload = f"""
        mutation {{
          {api_key}(annotation: {{
              flight: {{
                  id: {annotatable_flight_id}
              }},
              color: "{color}"
              labels: {json.dumps(labels) if labels else []}
              geometry: "{valid_geom}" }})
        {{
            id
            labels
            errors {{
                message
            }}
          }}
        }}"""
        response = self._make_request(payload=payload, api_key=api_key)
        return response

    def get_annotations(self, project_id: int, flight_code: Optional[str] = None,
                        api_limit_size: int = 1000) -> Generator[dict, None, None]:
        """Get annotations from a project

        Args:
            project_id: The id of the project to get annotations from
            flight_code: optionally filter by a flight
            api_limit_size: string annotation labels

        Returns:
            response from api call

        """
        if not flight_code:
            query_command = f"{{projectId: {project_id}}}"
        else:
            query_command = f"{{projectId: {project_id}, flightCode: \"{flight_code}\"}}"
        offset = 0
        while True:
            payload = f"""
                {{
                    listAnnotations(
                        max: {api_limit_size},
                        offset: {offset},
                        queryCommand: {query_command}
                    )
                    {{
                        results {{
                            flight {{
                                status
                                flight {{
                                    code
                                    tags
                                }}
                            }}
                            labels
                            geometry
                        }}
                        totalCount
                    }}
                }} """
            response = self.api.graphql_request(service='admin', query=payload)
            for result in response['data']['listAnnotations']['results']:
                geometry = shapely.geometry.shape(json.loads(result["geometry"]))
                yield {
                    'flight_code': result['flight']['flight']['code'],
                    'status': result['flight']['status'],
                    'labels': result['labels'],
                    'geometry': make_valid(geometry),
                    'tags': result['flight']['flight']['tags']
                }
            offset += api_limit_size
            if offset >= response['data']['listAnnotations']['totalCount']:
                break

    def download_annotations(self, project_id: int, output_path: Union[str, Path]):
        response = self.api.post('/annotations/export', ignore_json=True, params={
            "projectId": project_id,
            "flightCode": None,
            "labelIds": None
        })

        annotations_zipfile = zipfile.ZipFile(io.BytesIO(response.content))
        assert len(annotations_zipfile.filelist) == 1
        annotations = annotations_zipfile.read(annotations_zipfile.filelist[0])
        with open(output_path, 'wb') as fp:
            fp.write(annotations)

    def _make_request(self, payload: str, api_key: str):
        """Make actual call to graphql with provided payload.

        Args:
            payload: request payload
            api_key: name of api in use

        Returns:

        """
        response = self.api.graphql_request(service='admin', query=payload)
        errors = response["errors"] if 'errors' in response else response["data"][api_key].get("errors")
        if errors:
            raise Exception(f'Errors while making api request to {api_key}: {errors}')
        return response



"""Flight service API client."""

import logging
import urllib.parse
from typing import Optional

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from intelinair_utils.agmri_api import AgmriApi

# types:
AERIAL = 'aerial'
EQUIPMENT = 'equipment'
IMAGERY = 'imagery'

logger = logging.getLogger(__name__)


class FlightServiceApi:
    """API wrapper for flight service."""

    REQUEST_TIMEOUT = 180
    FLIGHT_SERVICE_URLS = {
        "ivesdale": "https://zt2rsi1flf.execute-api.us-east-1.amazonaws.com/api",
        "thomasboro": "https://ukfq5winf5.execute-api.us-east-1.amazonaws.com/api"
    }

    def __init__(self, environment: str, agmri_api: Optional[AgmriApi] = None, config_path: Optional[str] = None):
        """

        Args:
            environment: the env name
            agmri_api: initialized AgmriApi object
            config_path:  the optional path to agmri.cfg, if yoy want to use ssm, use ssm://<ssm param name>
        """
        self.environment = environment
        self.session = requests.Session()
        self.url_template = f"{self.FLIGHT_SERVICE_URLS[self.environment]}/{{}}"
        self._token = self._get_auth_token(agmri_api, config_path)
        self._headers = {
            'x-auth-token': self._token
        }
        retries = Retry(total=10, read=10, connect=10, backoff_factor=1,
                        allowed_methods=False,
                        status_forcelist=[409] + list(range(500, 600)))
        self.session.mount('https://', HTTPAdapter(max_retries=retries))

    def _get_auth_token(self, agmri_api: Optional[AgmriApi] = None, config_path: Optional[str] = None):
        if agmri_api is None:
            agmri_api = AgmriApi(self.environment, config_path=config_path)
        agmri_api.login()
        token = agmri_api.headers['processing']['X-Auth-Token']
        return token

    def _make_request(self, http_method, *args, ignore_json=False, **kwargs):
        response = http_method(*args, **kwargs, headers=self._headers, timeout=self.REQUEST_TIMEOUT)
        return response if ignore_json else response.json()

    def get_flight(self, flight_code: str, flight_type: str) -> dict:
        """
        Getting flight object.
        Args:
            flight_code: unique flight code.
            flight_type: either equipment or aerial.
        """
        url = self.url_template.format(f"flights/{flight_type}/{flight_code}")
        response = self._make_request(self.session.get, url)
        return response

    def create_flight(self, flight_type: str, flight_data: dict) -> dict:
        """
        Creates flight object with values passed through flight_data.
        Args:
            flight_type: either equipment or aerial.
            flight_data: properties of flight object to create.
        Returns:
            Created flight object.
        """
        url = self.url_template.format(f"flights/{flight_type}/create")
        response = self._make_request(self.session.post, url, json=flight_data)
        return response

    def update_flight(self, flight_type: str, flight_code: str, flight_data: dict):
        """
        Update flight object.
        Args:
            flight_type: either equipment or aerial.
            flight_code: code of flight to change.
            flight_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"flights/{flight_type}/{flight_code}/update")
        responce = self._make_request(self.session.post, url, json=flight_data)
        return responce

    def delete_flight(self, flight_type: str, flight_code: str):
        """
        Delete flight by code.
        Args:
            flight_type: either equipment or aerial.
            flight_code: code of flight to delete.
        """
        url = self.url_template.format(f"flights/{flight_type}/{flight_code}")
        self._make_request(self.session.delete, url, ignore_json=True)

    def add_flight_tags(self, flight_type: str, flight_code: str, tags: list):
        """
        Add flight tags
        Args:
            flight_type: either equipment or aerial.
            flight_code: code of flight to change.
            tags: list of tags to add
        Returns:
            list of tags after updates
        """
        url = self.url_template.format(f"flights/{flight_type}/{flight_code}/addTags")
        self._make_request(self.session.post, url, json=tags)

    def remove_flight_tags(self, flight_type: str, flight_code: str, tags: list):
        """
        Remove flight tags
        Args:
            flight_type: either equipment or aerial.
            flight_code: code of flight to change.
            tags: list of tags to Remove
        Returns:
            list of tags after updates
        """
        url = self.url_template.format(f"flights/{flight_type}/{flight_code}/removeTags")
        self._make_request(self.session.post, url, json=tags)

    def report_flights(self, flight_type, filters: dict):
        """
        Report flights based on input filters.
        Args:
            flight_type: either equipment or aerial.
            filters: key values, based on which flights will be reported.
                     additional ones are date_to, date_from, offset, limit
        Returns:
            list of flights reported by given filters.
        """
        query_params = urllib.parse.urlencode(filters)
        url = self.url_template.format(f"flights/report/{flight_type}?{query_params}")
        response = self._make_request(self.session.get, url)
        return response

    def get_imagery(self, imagery_code: str) -> dict:
        """
        Getting imagery object.
        Args:
            imagery_code: unique imagery code.
        """
        url = self.url_template.format(f"imageries/{imagery_code}")
        response = self._make_request(self.session.get, url)
        return response

    def create_imagery(self, imagery_data: dict) -> dict:
        """
        Creates imagery object with values passed through imagery_data.
        Args:
            imagery_data: properties of imagery object to create.
        Returns:
            Created imagery object.
        """
        url = self.url_template.format("imageries/create")
        response = self._make_request(self.session.post, url, json=imagery_data)
        return response

    def update_imagery(self, imagery_code: str, imagery_data: dict):
        """
        Update imagery object properties.
        Args:
            imagery_code: code of imagery to be changed.
            imagery_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"imageries/{imagery_code}/update")
        response = self._make_request(self.session.post, url, json=imagery_data)
        return response

    def delete_imagery(self, imagery_code: str):
        """
        Delete imagery by code.
        Args:
            imagery_code: code of imagery to delete.
        """
        url = self.url_template.format(f"imageries/{imagery_code}")
        self._make_request(self.session.delete, url, ignore_json=True)

    def get_tileset(self, tileset_type: str, code: str) -> dict:
        """
        Getting tileset object.
        Args:
            tileset_type: either equipment, aerial or imagery.
            code: flight/imagery object code for which tileset is returned.
        """
        url = self.url_template.format(f"tilesets/{tileset_type}/{code}")
        response = self._make_request(self.session.get, url)
        return response

    def create_tileset(self, tileset_type: str, code: str, tileset_data: dict) -> dict:
        """
        Creates tileset object with values passed through tileset_data.
        Args:
            tileset_type: either equipment, aerial or imagery.
            code: flight/imagery object code for which tileset is created.
            tileset_data: properties of tileset object to create.
        Returns:
            Created tileset object.
        """
        url = self.url_template.format(f"tilesets/{tileset_type}/{code}/create")
        response = self._make_request(self.session.post, url, json=tileset_data)
        return response

    def update_tileset(self, tileset_id: str, tileset_data: dict):
        """
        Update tileset object properties.
        Args:
            tileset_id: ID of tileset to be changed.
            tileset_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"/tilesets/update/{tileset_id}")
        response = self._make_request(self.session.post, url, json=tileset_data)
        return response

    def delete_tileset(self, tileset_id: int):
        """
        Delete tileset by ID.
        Args:
            tileset_id: ID of tileset to delete.
        """
        url = self.url_template.format(f"tilesets/delete/{tileset_id}")
        self._make_request(self.session.delete, url, ignore_json=True)



import json
import logging
import os
from typing import List, Union

import boto3

from intelinair_utils.os_utils import call_command

__all__ = ["attach_available_efses", "get_available_efses", "get_available_efses_from_s3", "get_efs_ip_address"]

logger = logging.getLogger(__name__)


def get_available_efses(
    env: str,
    efs_client: boto3.client,
    ec2_client: boto3.client,
    availability_zone: str
) -> Union[List, None]:
    """Find available efses based on environment

    Args:
        env (str): Environment of a machine
        efs_client (boto3.client): boto3 efs client
        ec2_client (boto3.client): boto3 ec2 client
        availability_zone (str): availability zone id

    Returns:
        Union[List, None]: Available efses, list of dicts with efs info
                            {
                                "id": "fs-id",
                                "name": "efs-name",
                                "ip_address": "ip-address",
                                "az": "availability-zone"
                            }
    """
    efs_to_attach = []
    efs_filesystems = efs_client.describe_file_systems()
    zone_id = ec2_client.describe_availability_zones(ZoneNames=[availability_zone])[
        "AvailabilityZones"][0]['ZoneId']
    for efs_filesystem in efs_filesystems['FileSystems']:
        efs_tags = parse_aws_tags(efs_filesystem['Tags'])
        if efs_tags.get('Environment', {}) == env and efs_tags.get('Service', {}) == 'pipeline':
            efs = {}
            efs['id'] = efs_filesystem['FileSystemId']
            efs['name'] = efs_filesystem['Name']
            ip_address = get_efs_ip_address(efs['id'], zone_id, efs_client, ec2_client)
            efs['ip_address'] = ip_address
            efs['az'] = availability_zone
            efs_to_attach.append(efs)
            break

    return efs_to_attach


def parse_aws_tags(tags: list) -> dict:
    """
    Simple method to parse AWS tags from list to dict.
    Args:
        tags: list of dicts of AWS object.
    Returns:
        tags in standard key value representation.
    """
    return {tag['Key']: tag['Value'] for tag in tags}


def get_available_efses_from_s3(s3_client: boto3.client, env: str, availability_zone: str):
    bucket = 'intelinair-infra'
    key = f'pipeline-metadata/efs/{env}/efs.json'
    obj = s3_client.get_object(Bucket=bucket, Key=key)
    efses = json.loads(obj['Body'].read().decode('utf-8'))

    efs_to_attach = []

    for efs in efses:
        if efs['az'] == availability_zone:
            efs_to_attach.append(efs)
    return efs_to_attach


def get_efs_ip_address(
    efs_id: str,
    zone_id: str,
    efs_client: boto3.client,
    ec2_client: boto3.client
) -> Union[ str, None]:
    """Finds ip address of efs based on efs id and availability zone id

    Args:
        efs_id (str): efs_id
        zone_id (str): availability zone id
        efs_client (boto3.client): boto3 efs client
        ec2_client (boto3.client): boto3 ec2 client

    Returns:
        str: The ip address of efs in the availability zone
    """

    ip_address = None
    mount_targets = efs_client.describe_mount_targets(FileSystemId=efs_id)['MountTargets']
    for mount_target in mount_targets:
        subnet_id = mount_target['SubnetId']
        subnets = ec2_client.describe_subnets(SubnetIds=[subnet_id])
        az_id = subnets['Subnets'][0]['AvailabilityZoneId']
        if az_id == zone_id:
            ip_address = mount_target['IpAddress']
            break
    return ip_address


def attach_available_efses(env: str, region: str, availability_zone: str) -> Union[List, None]:
    """Attaches available efses based on environment
       Filters efses based on env tag and attaches.

    Args:
        env (str): Environment of a machine
        region (str): Machine region
        availability_zone (str): Region's availability zone

    Returns:
        Union[List, None]: None in case of failure or List of attached efses
    """

    # if EFS isn't from prod than it's from dev
    if env != "prod":
        env = "dev"

    mount_points = []
    s3_client = boto3.client('s3')
    efs_to_attach = get_available_efses_from_s3(s3_client, env, availability_zone)

    for efs in efs_to_attach:
        mount_point = os.path.join("/mnt", efs['name'])

        if not os.path.exists(mount_point):
            os.system(f"sudo mkdir -p {mount_point}")

        # handling from different vpc case
        ip_address = efs['ip_address']
        cmd = f'echo "{ip_address} {efs["id"]}.efs.{region}.amazonaws.com" | sudo tee -a /etc/hosts'
        ret_code, error_txt, stderr = call_command(cmd, 'mount_efs', silent=True)
        if ret_code != 0:
            logger.error(stderr)
            return

        # mount efs
        cmd = f"sudo mount -t efs {efs['id']}:/ {mount_point}"
        ret_code, error_txt, stderr = call_command(cmd, 'mount_efs', silent=True)
        if ret_code != 0:
            logger.error(stderr)
            return
        os.system(f"sudo chown $USER:$USER {mount_point}")
        mount_points.append(mount_point)

    return mount_points



"""Planting service API wrapper class."""

import json
from datetime import datetime
from typing import Tuple, Union, List, Optional
from uuid import UUID

import shapely.geometry
import shapely.wkt
from pydantic import BaseModel, StrictInt, field_validator, ConfigDict

from intelinair_utils.api_wrapper import ApiWrapper


class PlantingServiceOperation(BaseModel):
    field_token: UUID
    season_id: StrictInt
    source: str  # SourceTypes enum: "USER", "MANUAL", "EQUIPMENT", "ANALYTICS"
    id: Optional[int] = None
    date: Optional[datetime] = None
    crop_types: Optional[List[str]] = None
    rate: Optional[int] = None
    geometry: Optional[Union[shapely.geometry.MultiPolygon, shapely.geometry.Polygon]] = None
    coverage_percentage: Optional[float] = None
    hybrid: Optional[str] = None
    active_date_range: Optional[Tuple[datetime, Union[datetime, None]]] = None
    is_active: Optional[bool] = None
    planted: Optional[bool] = None
    author: Optional[str] = None
    operation_id: Optional[UUID] = None
    date_created: Optional[datetime] = None
    replanted: Optional[bool] = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        json_encoders={
            shapely.geometry.MultiPolygon: lambda s: shapely.geometry.mapping(s) if s else None,
            shapely.geometry.Polygon: lambda s: shapely.geometry.mapping(s) if s else None,
        }
    )

    @field_validator('rate', mode='before')
    def coerce_fractional_rate_to_int(cls, value):
        """Coerce fractional rate to int."""
        if value is None:
            return None
        try:
            return int(float(value))
        except (ValueError, TypeError):
            raise ValueError(f"Invalid value for rate: {value}. Must be a number or a string representing a number.")

    @field_validator('geometry', mode='before')
    @classmethod
    def parse_geometry(cls, value):
        if value:
            if isinstance(value, shapely.geometry.MultiPolygon) or isinstance(value, shapely.geometry.Polygon):
                return value
            elif isinstance(value, str):
                return shapely.wkt.loads(value)
            elif isinstance(value, dict):
                return shapely.geometry.shape(value)
            raise ValueError(f"Could not parse geometry: {value}")

    def dict_jsonable(self):
        """Return dictionary which is serializable to json."""
        return json.loads(self.model_dump_json())


class PlantingServiceApi(ApiWrapper):
    """Planting service API class."""
    PLANTING_SERVICE_URLS = {
        "bellflower": "https://53h6dpom55.execute-api.us-east-1.amazonaws.com/api",
        "longview": "https://1hqmp8xg3l.execute-api.us-east-1.amazonaws.com/api/",
        "release": "https://ulliupxcdd.execute-api.us-east-1.amazonaws.com/api",
        "platform": "https://oczjs0gwu3.execute-api.us-east-1.amazonaws.com/api",
        "prod": "https://b3on5xa1fd.execute-api.us-east-1.amazonaws.com/api"
    }

    def __init__(self, environment: str, config_path: Optional[str] = None):
        super().__init__(environment, self._get_api_urls(), config_path=config_path)

    def _get_api_urls(self):
        """Overridden method for getting planting service API gateway URLs."""
        return self.PLANTING_SERVICE_URLS

    def get_all_planting_operations(
        self,
        field_token: Union[str, UUID],
        season_id: Union[int, str]
    ) -> List[PlantingServiceOperation]:
        """
        Get all planting operations of field during given season.
        Args:
            field_token: unique field identifier string.
            season_id: id of season for which field is checked.
        Returns:
            List of planting operations.
        """
        url = self.url_template.format(f"all-planting-operations/{field_token}/{season_id}")
        response = self._custom_chunked_get(url)
        return [PlantingServiceOperation(**resp_po) for resp_po in response]

    def get_active_planting_operation(
        self,
        field_token: Union[str, UUID],
        season_id: Union[int, str]
    ) -> PlantingServiceOperation:
        """
        Get current active(default) planting operation based on predefined priorities.
        Args:
            field_token: unique field identifier string.
            season_id: id of season for which field is checked.
        Returns:
            Default planting operation data.
        """
        url = self.url_template.format(f"planting-date/{field_token}/{season_id}")
        response = self._make_request(self.session.get, url)
        response.raise_for_status()
        return PlantingServiceOperation(**response.json())

    def is_planted(self, field_token: Union[str, UUID], season_id: Union[int, str]):
        """
        Check field is planted in given season or not.
        Args:
            field_token: unique field identifier string.
            season_id: id of season for which field is checked.
        Returns:
            Boolean response.
        """
        url = self.url_template.format(f"is-planted/{field_token}/{season_id}")
        response = self._make_request(self.session.get, url)
        response.raise_for_status()
        return response.json()

    def set_not_planted(self, field_token: Union[str, UUID], season_id: Union[int, str]) -> PlantingServiceOperation:
        """
        Set field as not planted.
        Args:
            field_token: unique field identifier string.
            season_id: id of season for which field is used.
        Returns:
            Created Planting Operation.
        """
        url = self.url_template.format(f"not-planted/{field_token}/{season_id}")
        response = self._make_request(self.session.post, url)
        response.raise_for_status()
        return PlantingServiceOperation(**response.json())

    def create_planting_operation(self, planting_op: PlantingServiceOperation) -> PlantingServiceOperation:
        """
        Creating single planting operation.
        Args:
            planting_op: data dictionary of new planting operation.
        """
        url = self.url_template.format("add-planting-operation")
        response = self._make_request(self.session.post, url, json=planting_op.dict_jsonable())
        response.raise_for_status()
        return PlantingServiceOperation(**response.json())

    def create_bulk_planting_operations(
        self, bulk_planting_ops: List[PlantingServiceOperation]
    ) -> List[PlantingServiceOperation]:
        """
        Creating planting operations in bulk.
        Args:
            bulk_planting_ops: list of dicts with new operations data.
        """
        url = self.url_template.format("add-bulk-planting-operation")
        response = self._make_request(self.session.post, url, json=[po.dict_jsonable() for po in bulk_planting_ops])
        response.raise_for_status()
        return [PlantingServiceOperation(**resp_po) for resp_po in response.json()]

    def update_planting_operation(
        self, updated_planting_operation: PlantingServiceOperation
    ) -> PlantingServiceOperation:
        """
        Update existing planting operation by ID.
        Args:
            updated_planting_operation: it should not include any data which will change planting operation itself
            e.g. (source, field_token etc.), so we that data is filtered.
        Returns:
            Updated planting operation.
        """
        url = self.url_template.format(f"planting-operations/{updated_planting_operation.id}")
        changeable_fields = ['geometry', 'rate', 'coverage', 'crop_types', 'replanted']
        payload = updated_planting_operation.dict_jsonable()
        payload = {key: value for key, value in payload.items() if key in changeable_fields}
        payload['geometry'] = json.dumps(payload['geometry'])
        response = self._make_request(self.session.patch, url, json=payload)
        response.raise_for_status()
        return PlantingServiceOperation(**response.json())

    def get_planting_priorities(self) -> dict:
        """
        Gets planting priorities.
        Returns:
            planting priorities.
        """
        url = self.url_template.format(f"priorities-config")
        response = self._make_request(self.session.get, url)
        response.raise_for_status()

        return response.json()



import logging
from typing import Optional

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from intelinair_utils.agmri_api import AgmriApi

logger = logging.getLogger(__name__)


class PipelineStatsApi:
    """API wrapper for pipeline-stats service, which implements get method for imagery and flight objects"""

    REQUEST_TIMEOUT = 180
    PIPELINE_STATS_URLS = {
        "prod": "https://0lnl943boa.execute-api.us-east-1.amazonaws.com/api/v1",
        "bellflower": "https://679l0mrptd.execute-api.us-east-1.amazonaws.com/api/v1",
        "release": "https://2ruql88p6k.execute-api.us-east-1.amazonaws.com/api/v1",
        "platform": "https://vxfckxgsnc.execute-api.us-east-1.amazonaws.com/api/v1"
    }

    def __init__(self, environment: str, config_path: Optional[str] = None):
        """

        Args:
            environment:  the env name
            config_path: the optional path to agmri.cfg, if yoy want to use ssm, use ssm://<ssm param name>
        """
        self.environment = environment
        self.config_path = config_path
        self.session = requests.Session()
        self.url_template = f"{self.PIPELINE_STATS_URLS[self.environment]}"
        self._token = self._get_auth_token()
        retries = Retry(total=10, read=10, connect=10, backoff_factor=1,
                        allowed_methods=False,
                        status_forcelist=[409] + list(range(500, 600)))
        self.session.mount('https://', HTTPAdapter(max_retries=retries))

    def _get_auth_token(self):
        agmri_api = AgmriApi(self.environment, config_path=self.config_path)
        agmri_api.ensure_authenticated()
        token = agmri_api.headers['processing']['X-Auth-Token']
        return token

    def _make_request(self, http_method, *args, **kwargs):
        response = http_method(*args, **kwargs, headers={'x-auth-token': self._token}, timeout=self.REQUEST_TIMEOUT)
        if response.status_code == 401:
            self._token = self._get_auth_token()
            return self._make_request(http_method, *args, **kwargs)
        return response

    def get_flight_stats(self, code, params=None, ignore_json=False):
        """Perform a GET request to the specified endpoint"""
        url = f"{self.url_template}/logs/flight/{code}"
        response = self._make_request(self.session.get, url, params=params)
        if ignore_json:
            return response
        return response.json()

    def get_imagery_stats(self, code, params=None, ignore_json=False):
        """Perform a GET request to the specified endpoint"""
        url = f"{self.url_template}/logs/imagery/{code}"
        response = self._make_request(self.session.get, url, params=params)
        if ignore_json:
            return response
        return response.json()

    def delete_stats(self, code, obj_type):
        """Perform a DELETE request to the specified endpoint"""

        valid_types = ("soil", "elevation", "equipment", "imagery", "flight")

        if obj_type in valid_types:
            url = f"{self.url_template}/logs/{obj_type}/{code}"
            response = self._make_request(self.session.delete, url)
            return response

        logger.warning("Please indicate valid object type")

    def get_stats_version(self, code, pipeline, processing_id, params=None):
        """Get stats version of a pipeline"""
        url = f"{self.url_template}/version/{pipeline}/{code}/{processing_id}"
        response = self._make_request(self.session.get, url, params=params)
        return response.json()



import logging

logger = logging.getLogger(__name__)

def log_after_first_attempt(retry_state):
    """Logs a generic retry message after the first attempt fails."""
    if retry_state.attempt_number > 1:
        func_name = retry_state.fn.__name__  # Get the function name being retried
        logger.info(f"Attempt {retry_state.attempt_number}: Failed to execute '{func_name}'. Retrying...")



"""Integration service API wrapper class."""

from typing import Union, Optional
from uuid import UUID

from intelinair_utils.api_wrapper import ApiWrapper


class IntegrationServiceApi(ApiWrapper):
    """Integration service API Wrapper."""
    INTEGRATION_SERVICE_URLS = {
        "longview": "https://api.longview.intelinair.dev/integration/leaf/operation",
        "release": "https://api.release.intelinair.dev/integration/leaf/operation",
        "prod": "https://api.ag-mri.intelinair.com/integration/leaf/operation"
    }

    def __init__(self, environment: str, config_path: Optional[str] = None):
        super().__init__(environment, self._get_api_urls(), config_path=config_path)

    def _get_api_urls(self):
        """Overridden method for getting integration service API URLs."""
        return self.INTEGRATION_SERVICE_URLS

    def get_operation_units(self, operation_id: Union[str, UUID]) -> dict:
        """
        Getting operation units from leaf integration.
        Args:
            operation_id: UUID of operation to get units for.
        """
        url = self.url_template.format("unit")
        url_params = {
            "operationId": operation_id
        }
        response = self._make_request(self.session.get, url, params=url_params)
        response.raise_for_status()
        return response.json()



import copy


class Map:
    def __init__(self, dictionary):
        d = {}
        for k, v in dictionary.items():
            d[k] = Map(v) if isinstance(v, dict) else v
        self.__dict__.update(d)

    def __repr__(self):
        return self.__dict__.__repr__()

    def __getitem__(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, value):
        if isinstance(value, dict):
            value = Map(value)
        self.__dict__[key] = value

    def __contains__(self, item):
        return item in self.__dict__

    def __copy__(self):
        return Map(self.__dict__.copy())

    def copy(self):
        return copy.copy(self)

    def to_dict(self):
        return self.__dict__.copy()


CONFIG_JSON_ = {
    "indices": {
        "scaling_coefficients": {
            "nir": 0.69,
            "red": 1,
            "green": 1.09,
        },
        "scale_rgb_brightness": True,
        "percentiles": {
            "red": {"low": 2, "high": 98},
            "green": {"low": 2, "high": 98},
            "blue": {"low": 1, "high": 99},
            "nir": {"low": 2, "high": 98},
        },
        "clip_ranges": {},
    },
    "veg_rows": {
        "fft_threshold": {
            "NDVI": 0.0001,
            "SciNDVI": 0.0004,
            "ExG": 0.015,
            "NIR": 0.00001,
            "GNDVI": 0.0001,
            "GREEN": 0.00001,
        },
        "veg_indices": ["NIR", "NDVI", "GREEN"],
        "max_distance": 150,
        "should_use_rgb_for_red_green": True
    },
    "change": {"veg_index": "NDVI", "compute_method": "diff", "blur_size": 450},
    "ndvi_anomaly": {
        "boundary_mask_erosion_in_cm": 500,
        "averaging_box_size": 450,
        "mean_range": [0.59, 0.61],
        "steps": [0.02, 0.06, 0.18],
        "soil_averaging_box_size": 15000,  # 150m
        "soil_ndvi_resize_factor": 4,  # should be integer
        "soil_correction_coefficient": 4,
        "locality_averaging_box_size": 0,  # global anomalies
        "locality_resize_factor": 8,  # should be integer
        "min_areas": {1: 1800, 2: 1400, 3: 1000},  # ~0.005 acres under ~10cm resolution
        "contour_config": {
            "tolerance": 0.1,
            "bezier": True,
        },
    },
    "gndvi_anomaly": {
        "boundary_mask_erosion_in_cm": 500,
        "averaging_box_size": 350,
        "mean_range": [0.59, 0.61],
        "steps": [0.03, 0.09, 0.27],
        "soil_averaging_box_size": 15000,  # 150m
        "soil_ndvi_resize_factor": 4,  # should be integer
        "soil_correction_coefficient": 0,  # no soil correction
        "locality_averaging_box_size": 0,  # global anomalies
        "locality_resize_factor": 8,  # should be integer
        "min_areas": {1: 1800, 2: 1400, 3: 1000},  # ~0.005 acres under ~10cm resolution
        "contour_config": {"tolerance": 0.1, "bezier": True},  # was 0.15
    },
    "thermal_local_anomaly": {
        "boundary_mask_erosion_in_cm": 500,
        "averaging_box_size": 450,
        "mean_range": [0.49, 0.51],
        "steps": [100, 300, 900],
        "soil_averaging_box_size": 15000,  # 150m
        "soil_ndvi_resize_factor": 4,  # should be integer
        "soil_correction_coefficient": 0,  # no soil correction
        "locality_averaging_box_size": 50000,  # 500m
        "locality_resize_factor": 10,  # should be integer
        "min_areas": {1: 1800, 2: 1800, 3: 1800},  # ~0.005 acres under ~10cm resolution
        "contour_config": {"tolerance": 0.1, "bezier": True},
    },
    "ndvi_anomaly_on_rows": {
        "boundary_mask_erosion_in_cm": 500,
        "apply_row_mask": True,
        "averaging_box_size": 450,
        "mean_range": [0.59, 0.61],
        "steps": [0.02, 0.06, 0.18],
        "soil_averaging_box_size": 15000,  # 150m
        "soil_ndvi_resize_factor": 4,  # should be integer
        "soil_correction_coefficient": 0,  # when rows are visible, soil effect is negligible
        "locality_averaging_box_size": 0,  # global anomalies
        "locality_resize_factor": 8,  # should be integer
        "min_areas": {1: 1800, 2: 1400, 3: 1000},  # ~0.005 acres under ~10cm resolution
        "contour_config": {"tolerance": 0.1, "bezier": True},
    },
    "gndvi_anomaly_on_rows": {
        "boundary_mask_erosion_in_cm": 500,
        "apply_row_mask": True,
        "averaging_box_size": 350,
        "mean_range": [0.59, 0.61],
        "steps": [0.03, 0.09, 0.27],
        "soil_averaging_box_size": 15000,  # 150m
        "soil_ndvi_resize_factor": 4,  # should be integer
        "soil_correction_coefficient": 0,  # no soil correction
        "locality_averaging_box_size": 0,  # global anomalies
        "locality_resize_factor": 8,  # should be integer
        "min_areas": {1: 1800, 2: 1400, 3: 1000},  # ~0.005 acres under ~10cm resolution
        "contour_config": {"tolerance": 0.1, "bezier": True},  # was 0.15
    },
    "weed_heatmap": {"cluster_distance": 1, "max_acceptable_area": None},  # in meters
    "stand_count": {
        "contour_config": {
            "tolerance": 0.1,  # was 0.5 or 0.3
            "bezier": True,
        }
    },
    "stand_count_no_rows": {
        "contour_config": {"tolerance": 0.1, "bezier": False}
    },
    "sparse_weed_detection": {"extra_args": {"extra_nodata_masks": ["waterway_detection:waterway_mask"]}},
    "equipment_application_dates": {"min_days_passed": 7, "max_days_passed": 21},
}

SEPERATED_CROPS = ["coffee"]


class AgmriConfig(Map):
    def __init__(self, crop_type=None):
        super().__init__(CONFIG_JSON_)
        self.indices.clip_ranges = {}
        if crop_type and crop_type in SEPERATED_CROPS:
            self.ndvi_anomaly = self.ndvi_anomaly_on_rows.copy()
            self.thermal_local_anomaly.apply_row_mask = True


class Sentinel2Config(AgmriConfig):
    def __init__(self, crop_type):
        super().__init__(crop_type)

        self.ndvi_anomaly.averaging_box_size = 0
        self.ndvi_anomaly.soil_correction_coefficient = 0
        self.ndvi_anomaly.locality_resize_factor = 1
        self.ndvi_anomaly.min_areas = {
            1: 180,  # The images are being upsampled by 10x for anomaly detection.
            2: 140,  # If the factor changes need to change here as well. Was 18 before
            3: 100,
        }


class MicasenseConfig(AgmriConfig):
    def __init__(self, crop_type):
        super().__init__(crop_type)

        self.ndvi_anomaly.min_areas = {1: 5400, 2: 4200, 3: 3000}

        self.ndvi_anomaly_on_rows.min_areas = {1: 5400, 2: 4200, 3: 3000}
        self.veg_rows.should_use_rgb_for_red_green = False

        self.indices.stretch_ranges_coefficients = {
            "red": {"low": 0, "high": 1},
            "green": {"low": 0, "high": 1},
            "blue": {"low": 0, "high": 1.15},
            "nir": {"low": 0, "high": 1},
        }


class PlanetPlanetScope(AgmriConfig):
    """
    Previously we used images from planet which required some preprocessing to be more human eye friendly.
    Now we get separate band which are already preprocessed by provider and will show these to the end user.
    Still, in case of old formatted input (missing red1, green1, blue1) we fal back to old way.
    At the same time, the analytics pipeline still uses the rgb channels which were used previously.
    """

    def __init__(self, crop_type):
        super().__init__(crop_type)

        self.indices.stretch_ranges_coefficients = {
            "red": {"low": 0.8, "high": 1},
            "green": {"low": 0.8, "high": 1},
            "blue": {"low": 0.8, "high": 1},
            "nir": {"low": 0.5, "high": 1},
        }

        self.indices.scale_rgb_brightness = True
        self.indices.percentiles.blue = {"low": 1, "high": 99}
        self.indices.clip_ranges = {"nir": (0, 5000), "red": (0, 1000000), "green": (0, 1000000), "blue": (0, 1000000)}

        self.ndvi_anomaly.averaging_box_size = 0
        self.ndvi_anomaly.soil_correction_coefficient = 0
        self.ndvi_anomaly.locality_resize_factor = 1
        self.ndvi_anomaly.min_areas = {1: 180, 2: 140, 3: 100}
        self.ndvi_anomaly.steps = [0.01, 0.05, 0.1]

        self.gndvi_anomaly.averaging_box_size = 0
        self.gndvi_anomaly.soil_correction_coefficient = 0
        self.gndvi_anomaly.locality_resize_factor = 1
        self.gndvi_anomaly.min_areas = {1: 180, 2: 140, 3: 100}
        self.gndvi_anomaly.steps = [0.01, 0.05, 0.1]


class AirbusConfig(AgmriConfig):
    def __init__(self, crop_type):
        super().__init__(crop_type)
        self.indices.scaling_coefficients = {"nir": 0.6, "red": 0.9, "green": 1}
        self.ndvi_anomaly.averaging_box_size = 0
        self.ndvi_anomaly.soil_correction_coefficient = 0
        self.ndvi_anomaly.locality_resize_factor = 1
        self.ndvi_anomaly.min_areas = {0: 200, 1: 180, 2: 1000, 3: 100}
        self.ndvi_anomaly.contour_config = {"tolerance": 0.1, "bezier": True}
        self.ndvi_anomaly.steps = [0.02, 0.06, 0.16]

        self.gndvi_anomaly.averaging_box_size = 0
        self.gndvi_anomaly.steps = [0.03, 0.06, 0.09, 0.15, 0.27]
        self.gndvi_anomaly.min_areas = {
            1: 1800,  # ~0.005 acres under ~10cm resolution
            2: 1400,
            3: 1000,
            4: 1000,
            5: 1000,
            6: 1000,
        }


CONFIGS_ = {
    "airbus_fake": {
        "default": AirbusConfig,
    },
    "airbus_pleiades": {
        "default": AirbusConfig,
    },
    "airbus_pleiades_1b": {
        "default": AirbusConfig,
    },
    "airbus_spot": {
        "default": AirbusConfig
    },
    "dji mavic 3m": {
        "default": MicasenseConfig
    },
    "sentinel2": {
        "default": Sentinel2Config,
    },
    "micasense": {
        "default": MicasenseConfig,
    },
    "micasense altum": {
        "default": MicasenseConfig,
    },
    "planet_planetscope": {
        "default": PlanetPlanetScope,
    }
}


def agmri_config_for_provider(provider, crop_type=None, config_type="default"):
    """
    Instantiate config for specified provider and specific type if provided
    :param provider:
    :param crop_type:
    :param config_type: Additional type to handle cases like different configs per provider
    :return: AgmriConfig instance
    """
    if provider not in CONFIGS_ or config_type not in CONFIGS_[provider]:
        print("provider {} / {} not found, returning base config".format(provider, config_type))
        return AgmriConfig(crop_type)
    return CONFIGS_[provider][config_type](crop_type)



import json
import logging
from functools import singledispatchmethod
from json import JSONDecodeError
from pathlib import Path
from subprocess import Popen
from typing import Dict, Union, Iterable, List

import yaml

logger = logging.getLogger(__file__)

PLUS: str = chr(43)  # "+"
LEAD: str = chr(45)  # "-"
BREAK: str = chr(32)  # " "
DOUBLE_LEAD: str = chr(45) * 2  # "--"


class PrettyPopen(Popen):

    def __init__(self, args: Union[Dict, str], **kwargs):
        """
        Example usage:
            ::arg::  arg currently the following structure:
                    +------------------------------------------------+
                     Dict[str: List[Tuple] # for glad_translate and exiftool
                     ## e.g. {"gdal_translate": [("-co", '"COMPRESS=DEFLATE"'),
                                                 ("-co", '"BIGTIFF=YES"'), ...]

        """

        super(PrettyPopen, self).__init__(args=self.process_arg(args), **kwargs)

    @classmethod
    def is_proper_dict(cls, args: Dict) -> bool:
        proper_dict = isinstance(args, dict) and len(args) == 1
        proper_values = (len(val) < 3 for val in args.values())
        return proper_values and proper_dict

    @staticmethod
    def merge_dicts(list_of_dicts: List[Dict]) -> Dict:
        """
        -- PyYAML reads a .yaml (as intended in the use case) into
            Dict[str: List[Dict]]
        -- JSON reads a .json into
            Dict[str: Dict]

        This function standardizes all of that.
        :param list_of_dicts:
        :return: the list of dicts merged into a single dict.
        """
        logger.debug(list_of_dicts)
        merged_dicts = dict(keyval for subdict in list_of_dicts for keyval in subdict.items())
        logger.debug(merged_dicts)
        return merged_dicts

    @classmethod
    def read_yaml(cls, file_path: str):
        with open(file_path, 'r') as stream:
            try:
                content = yaml.safe_load(stream)
                return {key: cls.merge_dicts(value) for key, value in content.items()}
            except yaml.YAMLError as exc:
                logger.exception(exc)
                raise exc

    @staticmethod
    def read_json(file_path: Union[str, bytes]):
        try:
            with open(file_path, 'r') as file:
                return json.load(file)

        except JSONDecodeError as exc:
            logger.exception(exc)
            raise exc

    def read_file(self, file: Path) -> Dict:
        """
        Args:
            :param file: Path to config file. e.g. ndvi_[something something].yaml
            :return: dict - the content of the file
        """
        logger.info(f"Reading subprocess command and arguments from: {file.as_posix()}")
        return getattr(self, f"read_{file.suffix[1:]}").__call__(file)

    @staticmethod
    def __unite(values: Iterable, joint: str = BREAK) -> str:
        unite = lambda values: joint.join(tuple(map(str, values)))
        united_arg_values = unite(values) if isinstance(values, (list, tuple)) else str(values)
        logger.debug(f"{values} > {united_arg_values}")
        return united_arg_values

    @staticmethod
    def __clean(long_string: str) -> str:
        """
        exiftool or gdal_translate commands
        have syntax of their own - hence this cleaning.
        """
        cleand_string = long_string.replace("= ", "=")
        logger.debug(cleand_string)
        return cleand_string

    def preprocess_args(self, arguments: Union[Dict, List]):

        # converting to list
        arguments = list(arguments.items()) if isinstance(arguments, dict) else arguments
        if isinstance(arguments, (list, tuple)):
            # Handling [**, ("arg_store") | ("arg_store", ), **] inputs
            arguments = [A if len(A) == 2 else tuple([*list(A), ""]) for A in arguments]
        return [(argument, self.__unite(value)) for argument, value in arguments]

    @singledispatchmethod
    def process_arg(self, args: Union[str, Path, Dict]):

        r"""Processes the arg as in a more handy manner.

            This can be called as
            .. function:: process_arg(str)
            .. function:: process_arg(dict)
            .. function:: process_arg(Path)

                        depending on the convenience.
            Returns:
                str: the command processible by subprocess.Popen

            Example::
                >>> args = {"subp.py": {"--arg_one": 102, "--arg_two": 205}}
                >>> PrettyPopen.process_arg(args, *, **)
                "python subp.py --arg_one 102 --arg_two 205"

                >>> args = "path_to_arg.json" # json:= {"subp.py": "--arg_one": 102, "--arg_two": 205}
                >>> PrettyPopen.process_arg(args, *, **)
                "python subp.py --arg_one 102 --arg_two 205"

                >>> args = "path_to_arg.yaml" # yaml:= subp.py:\n - --arg_one: 102\n - --arg_two: 205\n
                >>> PrettyPopen.process_arg(args, *, **)
                "python subp.py --arg_one 102 --arg_two 205"

                >>> args = "python subp.py --arg_one 102 --arg_two 205"
                >>> PrettyPopen.process_arg(args, *, **)
                "python subp.py --arg_one 102 --arg_two 205"
                """
        logger.debug(f"PrettyPopen arg passed as {type(args)}")
        raise NotImplementedError

    @process_arg.register
    def _(self, args: str):
        args_file = Path(args)
        if args_file.exists():
            return self.process_arg(args_file)
        return args

    @process_arg.register
    def _(self, args: Path) -> str:
        return self.process_arg(self.read_file(file=args))

    @process_arg.register
    def _(self, args: dict) -> str:
        assert self.is_proper_dict(args=args)
        try:
            # It is assumed that each cmd is about one process,
            # hence the length of the dict is 1.
            command = next(iter(args))
            arguments = args[command]
            arguments = self.preprocess_args(arguments)
            arguments = self.__unite([self.__unite(argument) for argument in arguments])
            return self.__unite([command, self.__clean(arguments)])
        except Exception as exc:
            logger.exception(exc)
            raise



import json
from typing import Any

import boto3
from botocore.exceptions import ClientError
from tenacity import retry, wait_exponential, stop_after_attempt, retry_if_exception_type


@retry(wait=wait_exponential(multiplier=10, exp_base=1.5), stop=stop_after_attempt(5),
       retry=retry_if_exception_type(ClientError), reraise=True)
def get_parameter_from_ssm(parameter_name: str) -> Any:
    """
    Retrieve a parameter from SSM store.
    Args:
        parameter_name: the name of the parameter to retrieve

    Returns:
        the parameter value
    """
    ssm_client = boto3.client('ssm')
    parameter = ssm_client.get_parameter(Name=parameter_name, WithDecryption=True)
    return parameter['Parameter']['Value']


def get_json_parameter_from_ssm(parameter_name: str) -> Any:
    """
    Retrieve a parameter from SSM store and parse it as JSON.
    Args:
        parameter_name: the name of the parameter to retrieve

    Returns:
        parameter value

    """
    value = get_parameter_from_ssm(parameter_name)
    try:
        parsed_value = json.loads(value)
    except Exception as e:
        raise e
    else:
        return parsed_value



"""API wrapper parent class for our microservices."""
import json
import logging
import os
import tempfile
from abc import ABC, abstractmethod
from typing import Dict, Optional

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from intelinair_utils.agmri_api import AgmriApi
from intelinair_utils.s3_utils import delete_s3_file, upload_s3_file

logger = logging.getLogger("api_wrapper")


class ApiWrapper(ABC):  # pylint: disable=too-few-public-methods
    """Parent class for our internal API wrapper clients."""
    REQUEST_TIMEOUT = 180
    MAX_UPLOAD_SIZE = 5000000  # ~5MB

    @abstractmethod
    def _get_api_urls(self):
        """Abstract method to ensure child classes had proper setup for API gateway urls."""
        return

    def get_s3_cache_paths(self) -> Dict[str, str]:
        """Return a dictionary of s3 cache paths for each environment supported."""
        return {}

    def __init__(self, environment: str, api_urls: dict, config_path: Optional[str] = None):
        """
        Args:
            environment: the env name
            api_urls: dict of url where the keys are env names and values are urls for making requests
            config_path: the optional path to agmri.cfg, if yoy want to use ssm, use ssm://<ssm param name>
        """
        self.environment = environment
        self.config_path = config_path
        self.session = requests.Session()
        self.url_template = f"{api_urls[self.environment]}/{{}}"
        self.s3_cache_path = self.get_s3_cache_paths().get(self.environment, None)
        self._token = self._get_auth_token()
        retries = Retry(
            total=10,
            read=10,
            connect=10,
            backoff_factor=1,
            allowed_methods=False,
            status_forcelist=[409] + list(range(500, 600))
        )
        self.session.mount('https://', HTTPAdapter(max_retries=retries))

    def _get_auth_token(self):
        agmri_api = AgmriApi(self.environment, config_path=self.config_path)
        agmri_api.ensure_authenticated()
        token = agmri_api.headers['processing']['X-Auth-Token']
        return token

    def _make_request(self, http_method, *args, attempt_authentication=True, headers=None, **kwargs):
        if not headers:
            headers = {}

        headers['x-auth-token'] = self._token
        response = http_method(*args, **kwargs, headers=headers, timeout=self.REQUEST_TIMEOUT)
        status_code = response.status_code

        if status_code == 401 and attempt_authentication:
            self._token = self._get_auth_token()
            return self._make_request(http_method, *args, attempt_authentication=False, headers=headers, **kwargs)

        return response

    def _upload_data_to_s3(self, json_data):
        with tempfile.NamedTemporaryFile('w', suffix='.json') as tmp_file:
            json.dump(json_data, tmp_file)
            tmp_file.flush()
            tmp_filename = os.path.basename(tmp_file.name)
            upload_s3_file(tmp_file.name, f'{self.s3_cache_path}/{tmp_filename}')
            return tmp_filename

    def _safe_delete_s3_file(self, filename):
        try:
            delete_s3_file(f'{self.s3_cache_path}/{filename}')
        except Exception as err:  # pylint: disable=broad-exception-caught
            logger.error(f'Unable to delete file {filename} from S3: {err}')

    def _make_large_body_request(self, http_method, *args, **kwargs):
        if self.s3_cache_path is not None and 'json' in kwargs:
            large_data_len = len(json.dumps(kwargs['json']))
            logger.debug(f'{self.__class__.__name__} request data length: {large_data_len:,}')
            if large_data_len > self.MAX_UPLOAD_SIZE:
                filename = self._upload_data_to_s3(kwargs['json'])
                try:
                    kwargs['json'] = {}
                    headers = kwargs.pop('headers', {})
                    headers['x-s3-request-body'] = f'{self.s3_cache_path}/{filename}'
                    response = self._make_request(http_method, headers=headers, *args, **kwargs)
                    return response
                finally:
                    self._safe_delete_s3_file(f'{self.s3_cache_path}/{filename}')
        response = self._make_request(http_method, *args, **kwargs)
        return response

    def _custom_chunked_get(self, *args, **kwargs):
        headers = kwargs.pop('headers', {})
        headers['x-chunked-output-byte'] = "0"
        chunks = []
        while True:
            response = self._make_request(self.session.get, *args, headers=headers, **kwargs)
            response.raise_for_status()
            chunks.append(response)
            if 'x-chunked-output-next-byte' in response.headers:
                headers['x-chunked-output-byte'] = response.headers['x-chunked-output-next-byte']
            else:
                break
        concatenated = ''.join([chunk.text for chunk in chunks])
        return json.loads(concatenated)



"""Module contains various utility functions for AWS"""
import json
import logging
import subprocess
import sys
import time
from typing import Dict, List

import boto3
from botocore.exceptions import ClientError
import requests
from tenacity import retry, wait_exponential, stop_after_attempt, retry_if_exception_type

from intelinair_utils.retry_utils import log_after_first_attempt


__all__ = [
    "get_ec2_tags", "create_ec2_tags", "delete_ec2_tags", "ec2_is_spot",
    "get_ec2_meta", "shut_down", "make_aws_request", "get_ec2_spot_price",
    "get_ec2_ondemand_price"
]

logger = logging.getLogger("ec2_utils")


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(8),
       reraise=True, after=log_after_first_attempt)
def get_ec2_tags(instance_id, region_name) -> Dict[str, str]:
    """Returns the EC2 tags for this instance."""

    ec2 = boto3.resource('ec2', region_name=region_name)
    i = ec2.Instance(id=instance_id)
    tags = dict([(i['Key'], i['Value']) for i in i.tags])

    return tags


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(8),
       reraise=True, after=log_after_first_attempt)
def create_ec2_tags(instance_id: str, tags: List[Dict[str, str]], region_name: str) -> None:
    ec2 = boto3.resource("ec2", region_name=region_name)
    ec2.create_tags(
        Resources=[instance_id],
        Tags=tags
    )


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(8),
       reraise=True, after=log_after_first_attempt)
def delete_ec2_tags(instance_id, region_name, tags) -> None:
    client = boto3.client("ec2", region_name=region_name)
    client.delete_tags(Resources=[instance_id], Tags=[{"Key": tag} for tag in tags])


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(8),
       reraise=True, after=log_after_first_attempt)
def ec2_is_spot(instance_id, region_name) -> bool:
    ec2 = boto3.resource('ec2', region_name=region_name)
    instance = ec2.Instance(id=instance_id)

    return True if instance.spot_instance_request_id is not None else False


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(8),
       reraise=True, after=log_after_first_attempt)
def get_ec2_meta() -> Dict[str, str]:
    # Get the instance ID and availability zone for this instance
    instance_id = requests.get(
        "http://169.254.169.254/latest/meta-data/instance-id", timeout=10).text
    logger.debug(f"AWS EC2 Instance ID: {instance_id}")
    availability_zone = requests.get(
        "http://169.254.169.254/latest/meta-data/placement/availability-zone", timeout=10).text
    logger.debug(f"AWS EC2 Availability Zone: {availability_zone}")
    instance_type = requests.get("http://169.254.169.254/latest/meta-data/instance-type", timeout=10).text
    logger.debug(f"AWS EC2 instance type: {instance_type}")
    meta = {
        "instance_id": instance_id,
        "instance_type": instance_type,
        "availability_zone": availability_zone
    }

    # Map from the availability zone to a region name
    try:
        int(availability_zone.split("-")[-1])
        region_name = availability_zone
    except ValueError:
        region_name = availability_zone[:-1]
    meta["region"] = region_name
    logger.debug(f"AWS EC2 Region: {region_name}")

    return meta


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(8),
       reraise=True, after=log_after_first_attempt)
def shut_down(in_ec2: bool = True) -> None:
    """Shut downs ec2 instance."""
    if in_ec2:
        # Power-off the server
        subprocess.run("sudo poweroff", shell=True)
    else:
        # this is local run, just exit
        sys.exit(1)


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(8), retry=retry_if_exception_type(ClientError),
       reraise=True, after=log_after_first_attempt)
def make_aws_request(function_name, kwargs, region_name=None):
    response = None
    if not region_name:
        region_name = boto3.DEFAULT_SESSION.region_name
    response = function_name(**kwargs)

    return response


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(8), retry=retry_if_exception_type(ClientError),
       reraise=True, after=log_after_first_attempt)
def get_ec2_spot_price(instance_type, availability_zone, operating_system):
    """Get EC2 instance spot price"""
    client = boto3.client('ec2')
    response = client.describe_spot_price_history(
        Filters=[{'Name': 'product-description', 'Values': [operating_system]}],
        AvailabilityZone=availability_zone,
        StartTime=time.time(),
        EndTime=time.time(),
        InstanceTypes=[instance_type])

    if len(response['SpotPriceHistory']) > 0:
        return float(response['SpotPriceHistory'][0]['SpotPrice'])
    else:
        return None


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(8), retry=retry_if_exception_type(ClientError),
       reraise=True, after=log_after_first_attempt)
def get_ec2_ondemand_price(instance_type, region, operating_system="Linux"):
    """
    Gets EC2 on-demand price based on provided parameters:
    "licenseModel": Checks to see if the target EC2 Instance price is for Bring your own license ,
    "preInstalledSw": This is for EC2 Instances that uses an EC2 Image (AMI) that has a preinstalled SQL Server.
    """
    client = boto3.client('pricing')
    response = client.get_products(
        ServiceCode='AmazonEC2',
        Filters=[
            {"Type": "TERM_MATCH", "Field": "regionCode", "Value": f"{region}"},
            {"Type": "TERM_MATCH", "Field": "operatingSystem", "Value": f"{operating_system}"},
            {"Type": "TERM_MATCH", "Field": "instanceType", "Value": f"{instance_type}"},
            {"Type": "TERM_MATCH", "Field": "marketoption", "Value": "OnDemand"},
            {"Type": "TERM_MATCH", "Field": "tenancy", "Value": "Shared"},
            {"Type": "TERM_MATCH", "Field": "preInstalledSw", "Value": "NA"},
            {"Type": "TERM_MATCH", "Field": "licenseModel", "Value": "No License required"},
            {"Type": "TERM_MATCH", "Field": "capacitystatus", "Value": "Used"}
        ]
    )

    if len(response['PriceList']) > 0:
        price = json.loads(response['PriceList'][0])
        for on_demand in price['terms']['OnDemand'].values():
            for price_dimensions in on_demand['priceDimensions'].values():
                price_value = price_dimensions['pricePerUnit']['USD']
                return float(price_value)
    else:
        return None



"""Collection of general utility functions"""
import logging
import multiprocessing
import os
from pathlib import Path
from typing import Tuple, Union
from urllib.parse import urlparse

import boto3
from botocore.exceptions import ClientError
from tenacity import retry, wait_exponential, retry_if_exception_type, stop_after_attempt

from intelinair_utils.retry_utils import log_after_first_attempt


__all__ = [
    'split_s3_path',
    'download_s3_file',
    's3_file_exists',
    'upload_s3_file',
    'download_from_s3',
    'upload_to_s3',
    'delete_s3_file',
    'get_list_of_s3_files'
]

s3_client = boto3.client('s3')
s3_resource = boto3.resource('s3')

logger = logging.getLogger(__name__)


def split_s3_path(s3_path: str) -> Tuple[str, str]:
    """Splits an s3 path into its bucket and key

    Args:
        s3_path: as s3 uri to split ex: 's3://your-bucket/your/key'

    Returns:
        bucket: the s3 bucket your object is located in
        key: the key of your object
    """
    if s3_path.startswith('https://s3.amazonaws.com/'):
        s3_path = s3_path.replace('https://s3.amazonaws.com/', 's3://')
    assert 's3://' == s3_path[0:5]
    path_parts = s3_path.replace("s3://", "").split("/")
    bucket = path_parts.pop(0)
    key = "/".join(path_parts)

    return bucket, key


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(3), retry=retry_if_exception_type(ClientError),
       reraise=True, after=log_after_first_attempt)
def download_s3_file(s3_path: str, output_path: Union[str, Path]) -> None:
    """Downloads the file from s3 to the given path."""
    logger.info(f'Downloading {s3_path} to {output_path}')

    output_path = Path(output_path).expanduser()
    bucket, key = split_s3_path(s3_path)

    if output_path.is_dir():
        output_path = output_path.joinpath(key.split('/')[-1])

    output_path.parent.mkdir(parents=True, exist_ok=True)
    s3_client.download_file(bucket, key, output_path.as_posix())


def download_from_s3(s3_path: str, local_path: Union[str, Path], num_processes: int = 8) -> None:
    """Downloads either a file or directory from s3"""
    local_path = Path(local_path)

    if s3_file_exists(s3_path) and not s3_path.endswith('/'):  # is actually a file
        return download_s3_file(s3_path, local_path.as_posix())

    def generate_args():
        bucket, prefix = split_s3_path(s3_path)
        bucket = s3_resource.Bucket(bucket)
        for obj in bucket.objects.filter(Prefix=prefix):
            if obj.key[-1] == '/':
                continue
            target = local_path.joinpath(obj.key.replace(prefix, '').lstrip('/'))
            yield f's3://{bucket.name}/{obj.key}', target.as_posix()

    if num_processes > 0:
        with multiprocessing.Pool(num_processes) as pool:
            pool.starmap(download_s3_file, generate_args())
    else:
        for args in generate_args():
            download_s3_file(*args)


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(3), retry=retry_if_exception_type(ClientError),
       reraise=True, after=log_after_first_attempt)
def upload_s3_file(local_path: str, s3_path: str) -> None:
    """Uploads the files from local_path to s3_path"""
    logger.info(f'Uploading {local_path} to {s3_path}')
    bucket, key = split_s3_path(s3_path)
    s3_client.upload_file(local_path, bucket, key)


@retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(3), retry=retry_if_exception_type(ClientError),
       reraise=True, after=log_after_first_attempt)
def delete_s3_file(s3_path: str) -> None:
    """Deletes the file from s3"""
    logger.info(f'Deleting {s3_path}')
    bucket, key = split_s3_path(s3_path)
    s3_client.delete_object(Bucket=bucket, Key=key)


def upload_to_s3(local_path: Union[str, Path], s3_path: str, num_processes: int = 8) -> None:
    """Uploads either a file or directory to s3"""
    local_path = Path(local_path)

    if local_path.is_file():
        return upload_s3_file(local_path.as_posix(), s3_path)

    bucket, key = split_s3_path(s3_path)

    def generate_args():
        for root, dirs, files in os.walk(local_path.as_posix()):

            for filename in files:
                # construct the full local path
                local_file = os.path.join(root, filename)

                # construct the full Dropbox path
                relative_path = os.path.relpath(local_file, local_path.as_posix())
                s3_file = os.path.join(key, relative_path)

                # relative_path = os.path.relpath(os.path.join(root, filename))
                yield local_file, f's3://{bucket}/{s3_file}'

    if num_processes > 0:
        with multiprocessing.Pool(num_processes) as pool:
            pool.starmap(upload_s3_file, generate_args())
    else:
        for args in generate_args():
            upload_s3_file(*args)


def s3_file_exists(s3_path: str) -> bool:
    """Check's if the file exists in a quick way"""
    bucket, key = split_s3_path(s3_path)
    try:
        s3_client.head_object(Bucket=bucket, Key=key)
        return True
    except ClientError:
        return False


def get_list_of_s3_files(remote_dir, list_of_file_extensions=None):
    files = []
    parts = urlparse(remote_dir, allow_fragments=False)
    bucket_name = parts.netloc
    search_prefix = parts.path.lstrip("/")
    bucket_obj = s3_resource.Bucket(bucket_name)

    for obj in bucket_obj.objects.filter(Prefix=search_prefix):
        if list_of_file_extensions and not obj.key.endswith(tuple(list_of_file_extensions)):
            continue
        files.append(os.path.join("s3://", bucket_name, obj.key))

    return files
